from typing import Any, Dict

from oml.interfaces.samplers import IBatchSampler
from oml.samplers.balance import BalanceSampler
from oml.samplers.category_balance import CategoryBalanceSampler
from oml.samplers.distinct_category_balance import DistinctCategoryBalanceSampler
from oml.utils.misc import TCfg, dictconfig_to_dict, remove_unused_kwargs

SAMPLERS_CATEGORIES_BASED = {
    "category_balance": CategoryBalanceSampler,
    "distinct_category_balance": DistinctCategoryBalanceSampler,
}

SAMPLERS_REGISTRY = {
    **SAMPLERS_CATEGORIES_BASED,  # type: ignore
    "balance": BalanceSampler,
}


def get_sampler(name: str, **kwargs: Dict[str, Any]) -> IBatchSampler:
    constructor = SAMPLERS_REGISTRY[name]
    kwargs = remove_unused_kwargs(kwargs, constructor)
    return constructor(**kwargs)  # type: ignore


def get_sampler_by_cfg(cfg: TCfg, **kwargs_runtime: Dict[str, Any]) -> IBatchSampler:
    cfg = dictconfig_to_dict(cfg)
    cfg["args"].update(kwargs_runtime)
    return get_sampler(name=cfg["name"], **cfg["args"])


__all__ = ["SAMPLERS_REGISTRY", "SAMPLERS_CATEGORIES_BASED", "get_sampler", "get_sampler_by_cfg"]



from abc import ABC
from typing import Any, Dict

from torch import Tensor, nn


class IExtractor(nn.Module, ABC):
    """
    Models have to inherit this interface to be comparable with the rest of the library.
    """

    pretrained_models: Dict[str, Any] = {}

    def extract(self, x: Tensor) -> Tensor:
        return self.forward(x)

    @property
    def feat_dim(self) -> int:
        """
        The only method that obligatory to implemented.
        """
        raise NotImplementedError()

    @classmethod
    def from_pretrained(cls, weights: str, **kwargs) -> "IExtractor":  # type: ignore
        """
        This method allows to download a pretrained checkpoint.
        The class field ``self.pretrained_models`` is the dictionary which keeps records of all the available
        checkpoints in the format, depending on implementation of a particular child of ``IExtractor``.
        As a user, you don't need to worry about implementing this method.

        Args:
            weights: A unique identifier (key) of a pretrained model information stored in
              a class field ``self.pretrained_models``.

        Returns: An instance of ``IExtractor``

        """
        if weights not in cls.pretrained_models:
            raise KeyError(
                f"There is no pretrained model {weights}. The existing ones are {list(cls.pretrained_models.keys())}."
            )

        extractor = cls(weights=weights, **cls.pretrained_models[weights]["init_args"], **kwargs)  # type: ignore
        return extractor


class IFreezable(ABC):
    """
    Models which can freeze and unfreeze their parts.
    """

    def freeze(self) -> None:
        """
        Function for freezing. You can use it to partially freeze a model.
        """
        raise NotImplementedError()

    def unfreeze(self) -> None:
        """
        Function for unfreezing. You can use it to unfreeze a model.
        """
        raise NotImplementedError()


class IPairwiseModel(nn.Module):
    """
    A model of this type takes two inputs, for example, two embeddings or two images.

    """

    def forward(self, x1: Any, x2: Any) -> Tensor:
        """

        Args:
            x1: The first input.
            x2: The second input.

        """
        raise NotImplementedError()

    def predict(self, x1: Any, x2: Any) -> Tensor:
        """
        While ``self.forward()`` is called during training, this method is called during
        inference or validation time. For example, it allows application of some activation,
        which was a part of a loss function during the training.

        Args:
            x1: The first input.
            x2: The second input.

        """
        raise NotImplementedError()


__all__ = ["IExtractor", "IFreezable", "IPairwiseModel"]



from typing import Optional, Union

import albumentations as albu
import torchvision.transforms as t

from oml.utils.images.images import TImReader, imread_cv2, imread_pillow

TTransforms = Union[albu.Compose, t.Compose]


def get_im_reader_for_transforms(transforms: Optional[TTransforms]) -> TImReader:
    if isinstance(transforms, t.Compose):
        return imread_pillow
    else:
        return imread_cv2


__all__ = ["TTransforms", "get_im_reader_for_transforms"]



from abc import ABC, abstractmethod
from typing import Iterator, List


class IBatchSampler(ABC):
    """
    We introduce our interface instead of using the default BatchSampler from Torch,
    because the last one is just a wrapper for the sequential sampler, which is not
    convenient for our purposes.

    """

    @abstractmethod
    def __len__(self) -> int:
        """
        Returns:
             The number of batches in an epoch

        """
        raise NotImplementedError()

    @abstractmethod
    def __iter__(self) -> Iterator[List[int]]:
        """
        Returns:
            Iterator contains indices for the batches

        """
        raise NotImplementedError()


__all__ = ["IBatchSampler"]



from typing import List

import numpy as np
from torch import Tensor

from oml.interfaces.miners import ITripletsMinerInBatch, TTripletsIds
from oml.utils.misc import find_value_ids
from oml.utils.misc_torch import pairwise_dist


class HardTripletsMiner(ITripletsMinerInBatch):
    """
    This miner selects the hardest triplets based on the distances between the features:

    - The hardest `positive` sample has the `maximal` distance to the anchor sample

    - The hardest `negative` sample has the `minimal` distance to the anchor sample

    """

    def _sample(self, features: Tensor, labels: List[int]) -> TTripletsIds:
        """
        This method samples the hardest triplets inside the batch.

        Args:
            features: Features with the shape of ``[batch_size, feature_size]``
            labels: Labels with the size of ``batch_size``

        Returns:
            The batch of the triplets in the order below:
            ``(anchor, positive, negative)``

        """
        assert features.shape[0] == len(labels)

        features = features.clone().detach()

        dist_mat = pairwise_dist(x1=features, x2=features, p=2)

        ids_anchor, ids_pos, ids_neg = self._sample_from_distmat(distmat=dist_mat, labels=labels)

        return ids_anchor, ids_pos, ids_neg

    @staticmethod
    def _sample_from_distmat(distmat: Tensor, labels: List[int]) -> TTripletsIds:
        """
        This method samples the hardest triplets based on the given
        distances matrix. It chooses each sample in the batch as an
        anchor and then finds the hardest positive and negative pair.

        Args:
            distmat: Matrix with the shape of ``[batch_size, batch_size]``
            labels: Labels with the size of ``batch_size``

        Returns:
            The batch of the triplets with the size of ``batch_size``, order is the following:
                ``(anchor, positive, negative)``

        """
        ids_all = set(range(len(labels)))

        ids_anchor, ids_pos, ids_neg = [], [], []

        for i_anch, label in enumerate(labels):
            ids_label = set(find_value_ids(it=labels, value=label))

            ids_pos_cur = np.array(list(ids_label - {i_anch}), int)
            ids_neg_cur = np.array(list(ids_all - ids_label), int)

            i_pos = ids_pos_cur[distmat[i_anch, ids_pos_cur].argmax()]
            i_neg = ids_neg_cur[distmat[i_anch, ids_neg_cur].argmin()]

            ids_anchor.append(i_anch)
            ids_pos.append(i_pos)
            ids_neg.append(i_neg)

        return ids_anchor, ids_pos, ids_neg


__all__ = ["HardTripletsMiner"]



from collections import defaultdict
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from torch import BoolTensor, FloatTensor, LongTensor, Tensor, isin
from tqdm.auto import tqdm

from oml.const import OVERALL_CATEGORIES_KEY
from oml.utils.misc import check_if_nonempty_positive_integers
from oml.utils.misc_torch import PCA

TMetricsDict = Dict[str, Any]
TCategories = Union[LongTensor, np.ndarray]


def calc_retrieval_metrics(
    retrieved_ids: Sequence[LongTensor],
    gt_ids: Sequence[LongTensor],
    query_categories: Optional[TCategories] = None,
    cmc_top_k: Tuple[int, ...] = (5,),
    precision_top_k: Tuple[int, ...] = (5,),
    map_top_k: Tuple[int, ...] = (5,),
    reduce: bool = True,
    verbose: bool = True,
) -> TMetricsDict:
    """
    Function to compute different retrieval metrics.

    Args:
        retrieved_ids: First gallery indices retrieved for every query with the size of ``n_query``.
            Every index is within the range ``(0, n_gallery - 1)``.
        gt_ids: Gallery indices relevant to every query with the size of ``n_query``.
            Every element is within the range ``(0, n_gallery - 1)``
        query_categories: Categories of queries with the size of ``n_query`` to compute metrics for each category.
        cmc_top_k: Values of ``k`` to calculate ``cmc@k`` (`Cumulative Matching Characteristic`)
        precision_top_k: Values of ``k`` to calculate ``precision@k``
        map_top_k: Values of ``k`` to calculate ``map@k`` (`Mean Average Precision`)
        reduce: If ``False`` return metrics for each query without averaging
        verbose: Set ``True`` to make the function verbose.

    Returns:
        Metrics dictionary.

    """
    assert len(retrieved_ids) == len(gt_ids)
    assert (query_categories is None) or (len(query_categories) == len(retrieved_ids))

    # let's mark every correctly retrieved item as True and vice versa
    gt_tops = tuple([isin(r, g).bool() for r, g in zip(retrieved_ids, gt_ids)])
    n_gts = [len(ids) for ids in gt_ids]

    metrics: TMetricsDict = defaultdict(dict)

    if cmc_top_k:
        cmc = calc_cmc(gt_tops, n_gts, cmc_top_k, verbose=verbose)
        metrics["cmc"] = dict(zip(cmc_top_k, cmc))

    if precision_top_k:
        precision = calc_precision(gt_tops, n_gts, precision_top_k, verbose=verbose)
        metrics["precision"] = dict(zip(precision_top_k, precision))

    if map_top_k:
        map_ = calc_map(gt_tops, n_gts, map_top_k, verbose=verbose)
        metrics["map"] = dict(zip(map_top_k, map_))

    if query_categories is not None:
        metrics_cat = {c: take_unreduced_metrics_by_mask(metrics, query_categories == c) for c in query_categories}
        metrics = {OVERALL_CATEGORIES_KEY: metrics, **metrics_cat}

    if reduce:
        metrics = reduce_metrics(metrics)

    return metrics


def calc_topological_metrics(
    embeddings: Tensor, pcf_variance: Tuple[float, ...], categories: Optional[TCategories] = None, verbose: bool = False
) -> TMetricsDict:
    """
    Function to evaluate different topological metrics.

    Args:
        embeddings: Embeddings matrix with the shape of ``[n_embeddings, embeddings_dim]``.
        categories: Categories of embeddings to compute category wise metrics.
        pcf_variance: Values in range [0, 1]. Find the number of components such that the amount
                      of variance that needs to be explained is greater than the percentage specified
                      by ``pcf_variance``.
        verbose: Set ``True`` to see a progress bar.

    Returns:
        Metrics dictionary.

    """
    assert (categories is None) or (len(categories) == len(embeddings))

    metrics: TMetricsDict = defaultdict(dict)

    if pcf_variance:
        main_components = calc_pcf(embeddings, pcf_variance)
        metrics["pcf"] = dict(zip(pcf_variance, main_components))

    if pcf_variance and (categories is not None):
        categories_unq = np.unique(categories)
        data = tqdm(categories_unq, desc="Topologic metrics on different categories") if verbose else categories_unq
        metrics_cat = {
            c: calc_topological_metrics(embeddings[categories == c], pcf_variance, categories=None) for c in data
        }
        metrics = {OVERALL_CATEGORIES_KEY: metrics, **metrics_cat}

    return metrics


def reduce_metrics(metrics_to_reduce: TMetricsDict) -> TMetricsDict:
    output: TMetricsDict = {}

    for k, v in metrics_to_reduce.items():
        if isinstance(v, (Tensor, np.ndarray)):
            output[k] = v.mean()
        elif isinstance(v, (float, int)):
            output[k] = v
        else:
            output[k] = reduce_metrics(v)  # type: ignore

    return output


def take_unreduced_metrics_by_mask(metrics: TMetricsDict, mask: BoolTensor) -> TMetricsDict:
    output: TMetricsDict = {}

    for k, v in metrics.items():
        if isinstance(v, Tensor):
            output[k] = v[mask] if v.numel() > 1 else v
        elif isinstance(v, (float, int)):
            output[k] = v
        else:
            output[k] = take_unreduced_metrics_by_mask(v, mask)  # type: ignore

    return output


def calc_cmc(
    gt_tops: Sequence[BoolTensor], n_gts: List[int], top_k: Tuple[int, ...], verbose: bool = False
) -> List[FloatTensor]:
    """
    Function to compute Cumulative Matching Characteristics (CMC) at cutoffs ``top_k``.

    ``cmc@k`` for a given query equals to 1 if there is at least 1 instance related to this query in top ``k``
    gallery instances sorted by distances to the query, and 0 otherwise.
    The final ``cmc@k`` could be obtained by averaging the results calculated for each query.

    Args:
        gt_tops: Indicators that show if retrievied items are correct or not:
            ``gt_tops[i][j]`` is ``True`` if ``j``-th gallery item is related to the ``i``-th query item.
        n_gts: Number of existing ground truths for every query.
        top_k: Values of ``k`` to calculate ``cmc@k``.
        verbose: Set ``True`` to see progress bar.

    Returns:
        List of ``cmc@k`` tensors computed for every query.

    .. math::
        \\textrm{cmc}@k = \\begin{cases}
        1, & \\textrm{if top-}k \\textrm{ ranked gallery samples include an output relevant to the query}, \\\\
        0, & \\textrm{otherwise}.
        \\end{cases}

    Example:
        >>> gt_tops = [
        ...     BoolTensor([1, 0]),
        ...     BoolTensor([0, 1, 1]),
        ...     BoolTensor([0, 0]),
        ...     BoolTensor([])
        ... ]
        >>> n_gts = [2, 2, 1, 0]
        >>> calc_cmc(gt_tops, n_gts, top_k=(1, 2))
        [tensor([1., 0., 0., 1.]), tensor([1., 1., 0., 1.])]
    """
    check_if_nonempty_positive_integers(top_k, "top_k")

    def cmc_single(is_correct: BoolTensor, n_gt: int, k_: int) -> float:
        if n_gt == 0 and len(is_correct) == 0:
            return 1.0
        elif n_gt > 0 and len(is_correct) == 0:
            return 0.0
        else:
            value = float(is_correct[:k_].any())
            return value

    cmc = []
    for k in top_k:
        items = tqdm(zip(gt_tops, n_gts), desc=f"CMC@{k}", total=len(gt_tops)) if verbose else zip(gt_tops, n_gts)
        cmc.append(FloatTensor([cmc_single(gts, n_gt, k) for gts, n_gt in items]))

    return cmc


def calc_precision(
    gt_tops: Sequence[BoolTensor], n_gts: List[int], top_k: Tuple[int, ...], verbose: bool = False
) -> List[FloatTensor]:
    """
    Function to compute Precision at cutoffs ``top_k``.

    ``precision@k`` for a given query is a fraction of the relevant gallery instances among the top ``k`` instances
    sorted by distances from the query to the gallery.
    The final ``precision@k`` could be obtained by averaging the results calculated for each query.

    Args:
        gt_tops: Indicators that show if retrievied items are correct or not:
            ``gt_tops[i][j]`` is ``True`` if ``j``-th gallery item is related to the ``i``-th query item.
        n_gts: Number of existing ground truth for every query.
        top_k: Values of ``k`` to calculate ``precision@k``.
        verbose: Set ``True`` to see progress bar.

    Returns:
        List of ``precision@k`` tensors computed for every query.

    Given a list :math:`g=[g_1, \\ldots, g_k]` of ground truth top :math:`k` closest elements from the gallery to
    a given query (:math:`g_i` is 1 if :math:`i`-th element from the gallery is relevant to the query and 0 otherwise),
    and the total number of relevant elements from the gallery :math:`n`, the :math:`\\textrm{precision}@k`
    for the query is defined as

    .. math::
        \\textrm{precision}@k = \\frac{1}{\\min{\\left(k, n\\right)}}\\sum\\limits_{i = 1}^k g_i

    It's worth mentioning that OML version of :math:`\\textrm{precision}@k` differs from the commonly used by
    the denominator of the fraction. The OML version takes into account the total amount of relevant elements in
    the gallery, so it will not penalize the ideal model if :math:`n < k`.

    For instance, let :math:`n = 3` and :math:`g = [1, 1, 1, 0, 0]`. Then by using the common definition of
    :math:`\\textrm{precision}@k` we get

    .. math::
        \\begin{align}
            \\textrm{precision}@1 &= \\frac{1}{1}, \\textrm{precision}@2 = \\frac{2}{2},
            \\textrm{precision}@3 = \\frac{3}{3}, \\\\
            \\textrm{precision}@4 &= \\frac{3}{4}, \\textrm{precision}@5 = \\frac{3}{5},
            \\textrm{precision}@6 = \\frac{3}{6} \\\\
        \\end{align}

    But with OML definition of :math:`\\textrm{precision}@k` we get


    .. math::
        \\begin{align}
            \\textrm{precision}@1 &= \\frac{1}{1}, \\textrm{precision}@2 = \\frac{2}{2},
            \\textrm{precision}@3 = \\frac{3}{3} \\\\
            \\textrm{precision}@4 &= \\frac{3}{3}, \\textrm{precision}@5 = \\frac{3}{3},
            \\textrm{precision}@6 = \\frac{3}{3} \\\\
        \\end{align}

    See:
        `Evaluation measures (information retrieval). Precision@k`_


        .. _`Evaluation measures (information retrieval). Precision@k`:
            https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Precision_at_k

    Example:

        >>> gt_tops = [
        ...     BoolTensor([1, 0]),
        ...     BoolTensor([0, 1, 1]),
        ...     BoolTensor([0, 0]),
        ...     BoolTensor([])
        ... ]
        >>> n_gts = [2, 3, 5, 2]
        >>> calc_precision(gt_tops, n_gts, top_k=(1, 2))
        [tensor([1., 0., 0., 0.]), tensor([0.5000, 0.5000, 0.0000, 0.0000])]

    """
    check_if_nonempty_positive_integers(top_k, "top_k")

    def precision_single(is_correct: BoolTensor, n_gt: int, k_: int) -> float:
        if n_gt == 0 and len(is_correct) == 0:
            return 1.0
        elif n_gt > 0 and len(is_correct) == 0:
            return 0.0
        else:
            k_ = min(k_, len(is_correct))
            value = torch.cumsum(is_correct, dim=0)[k_ - 1] / min(n_gt, k_)
            return float(value)

    precision = []
    for k in top_k:
        items = tqdm(zip(gt_tops, n_gts), desc=f"Precision@{k}", total=len(n_gts)) if verbose else zip(gt_tops, n_gts)
        precision.append(FloatTensor([precision_single(gts, n_gt, k) for gts, n_gt in items]))

    return precision


def calc_map(
    gt_tops: Sequence[BoolTensor], n_gts: List[int], top_k: Tuple[int, ...], verbose: bool = False
) -> List[FloatTensor]:
    """
    Function to compute Mean Average Precision (MAP) at cutoffs ``top_k``.

    ``map@k`` for a given query is the average value of the ``precision`` considered as a function of the ``recall``.
    The final ``map@k`` could be obtained by averaging the results calculated for each query.

    Args:
        gt_tops: Indicators that show if retrievied items are correct or not:
            ``gt_tops[i][j]`` is ``True`` if ``j``-th gallery item is related to the ``i``-th query item.
        n_gts: Number of existing ground truth for every query.
        top_k: Values of ``k`` to calculate ``map@k``.
        verbose: Set ``True`` to see progress bar.

    Returns:
        List of ``map@k`` tensors computed for every query.

    Given a list :math:`g=[g_1, \\ldots, g_k]` of ground truth top :math:`k` closest elements from the gallery to
    a given query (:math:`g_i` is 1 if :math:`i`-th element from the gallery is relevant to the query and 0 otherwise),
    and the total number of relevant elements from the gallery :math:`n`, the :math:`\\textrm{map}@k`
    for the query is defined as

    .. math::
        \\begin{split}
        \\textrm{map}@k &=
        \\frac{1}{n_k}\\sum\\limits_{i = 1}^k
        \\frac{n_i}{i} \\times \\textrm{rel}(i)
        \\end{split}

    where :math:`\\textrm{rel}(i)` is 1 if :math:`i`-th element from the top :math:`i` closest
    elements from the gallery to the query is relevant to the query, and 0 otherwise;
    and :math:`n_i = \\sum\\limits_{j = 1}^{i}g_j`, which is the number of the relevant predictions
    among the first :math:`i` outputs.

    See:

        `Evaluation measures (information retrieval). Mean Average Precision`_

        `Mean Average Precision (MAP) For Recommender Systems`_

    .. _`Evaluation measures (information retrieval). Mean Average Precision`:
        https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Mean_average_precision

    .. _`Mean Average Precision (MAP) For Recommender Systems`:
        https://sdsawtelle.github.io/blog/output/mean-average-precision-MAP-for-recommender-systems.html

    Example:
        >>> gt_tops = [
        ...    BoolTensor([1, 0]),
        ...    BoolTensor([0, 1]),
        ...    BoolTensor([0, 0, 0, 0]),
        ...    BoolTensor([])
        ... ]
        >>> n_gts = [1, 1, 2, 0]
        >>> calc_map(gt_tops, n_gts, top_k=(1, 2))
        [tensor([1., 0., 0., 1.]), tensor([1.0000, 0.5000, 0.0000, 1.0000])]
    """
    check_if_nonempty_positive_integers(top_k, "top_k")

    def map_single(is_correct: BoolTensor, n_gt: int, k_: int) -> float:
        if n_gt == 0 and len(is_correct) == 0:
            return 1.0
        elif n_gt > 0 and len(is_correct) == 0:
            return 0.0
        else:
            k_ = min(k_, len(is_correct))
            correct_preds = torch.cumsum(is_correct, dim=0).float()
            positions = torch.arange(1, k_ + 1).to(correct_preds.device)
            n_k = correct_preds[k_ - 1].clone()
            n_k[n_k < 1] = torch.inf  # hack to avoid zero division
            value = torch.sum((correct_preds[:k_] / positions) * is_correct[:k_], dim=0) / n_k
            return float(value)

    map_ = []
    for k in top_k:
        items = tqdm(zip(gt_tops, n_gts), total=len(gt_tops), desc=f"MAP@{k}") if verbose else zip(gt_tops, n_gts)
        map_.append(FloatTensor([map_single(is_correct, n_gt, k) for is_correct, n_gt in items]))

    return map_


def calc_fnmr_at_fmr(pos_dist: np.ndarray, neg_dist: np.ndarray, fmr_vals: Tuple[float, ...] = (0.1,)) -> FloatTensor:
    """
    Function to compute False Non Match Rate (FNMR) value when False Match Rate (FMR) value
    is equal to ``fmr_vals``.

    The metric calculates the quantile of positive distances higher than a given :math:`q`-th quantile
    of negative distances.

    Args:
        pos_dist: Distances between relevant samples.
        neg_dist: Distances between non-relevant samples.
        fmr_vals: Values of ``fmr`` (measured in quantiles) to compute the corresponding ``fnmr``.
                  For example, if ``fmr_values`` is (0.2, 0.4) we will calculate ``fnmr@fmr=0.2`` and ``fnmr@fmr=0.4``

    Returns:
        Tensor of ``fnmr@fmr`` values.

    Given a vector of :math:`N` distances between relevant samples, :math:`u`,
    the false non-match rate (:math:`\\textrm{FNMR}`) is computed as the proportion of :math:`u` below some threshold,
    :math:`T`:

    .. math::

        \\textrm{FNMR}(T) = \\frac{1}{N}\\sum\\limits_{i = 1}^{N}H\\left(u_i - T\\right) =
        1 - \\frac{1}{N}\\sum\\limits_{i = 1}^{N}H\\left(T - u_i\\right)

    where :math:`H(x)` is the unit step function, and :math:`H(0)` taken to be :math:`1`.

    Similarly, given a vector of :math:`N` distances between non-relevant samples, :math:`v`,
    the false match rate (:math:`\\textrm{FMR}`) is computed as the proportion of :math:`v` above some threshold,
    :math:`T`:

    .. math::

        \\textrm{FMR}(T) = 1 - \\frac{1}{N}\\sum\\limits_{i = 1}^{N}H\\left(v_i - T\\right) =
        \\frac{1}{N}\\sum\\limits_{i = 1}^{N}H\\left(T - v_i\\right)

    Given some interesting false match rate values :math:`\\textrm{FMR}_k` one can find thresholds :math:`T_k`
    corresponding to :math:`\\textrm{FMR}` measurements

    .. math::

        T_k = Q_v\\left(\\textrm{FMR}_k\\right)

    where :math:`Q` is the quantile function, and evaluate the corresponding values of
    :math:`\\textrm{FNMR}@\\textrm{FMR}\\left(T_k\\right) \\stackrel{\\text{def}}{=} \\textrm{FNMR}\\left(T_k\\right)`.


    See:

        `Biometrics Performance`_

        `BIOMETRIC RECOGNITION: A MODERN ERA FOR SECURITY`_

    .. _`Biometrics Performance`:
        https://en.wikipedia.org/wiki/Biometrics#Performance

    .. _`BIOMETRIC RECOGNITION: A MODERN ERA FOR SECURITY`:
        https://www.researchgate.net/publication/50315614_BIOMETRIC_RECOGNITION_A_MODERN_ERA_FOR_SECURITY


    Example:
        >>> pos_dist = np.array([0, 0, 1, 1, 2, 2, 5, 5, 9, 9])
        >>> neg_dist = np.array([3, 3, 4, 4, 6, 6, 7, 7, 8, 8])
        >>> calc_fnmr_at_fmr(pos_dist, neg_dist, fmr_vals=(0.1, 0.5))
        tensor([0.4000, 0.2000])

    """
    _check_if_in_range(fmr_vals, 0, 1, "fmr_vals")
    thresholds = np.quantile(neg_dist, fmr_vals)  # we use numpy because it can take bigger input
    fnmr_at_fmr = (pos_dist[None, :] >= thresholds[:, None]).sum(axis=1) / len(pos_dist)
    return FloatTensor(fnmr_at_fmr)


def calc_fnmr_at_fmr_by_distances(
    pos_dist: np.ndarray, neg_dist: np.ndarray, fmr_vals: Tuple[float, ...]
) -> TMetricsDict:
    metrics: TMetricsDict = dict()

    if fmr_vals:
        fnmr_at_fmr = calc_fnmr_at_fmr(pos_dist, neg_dist, fmr_vals)
        metrics["fnmr@fmr"] = dict(zip(fmr_vals, fnmr_at_fmr))

    return metrics


def calc_pcf(embeddings: Tensor, pcf_variance: Tuple[float, ...]) -> List[Tensor]:
    """
    Function estimates the Principal Components Fraction (PCF) of embeddings using Principal Component Analysis.
    The metric is defined as a fraction of components needed to explain the required variance in data.

    Args:
        embeddings: Embeddings matrix with the shape of ``[n_embeddings, embeddings_dim]``.
        pcf_variance: Values in range [0, 1]. Find the number of components such that the amount
                      of variance that needs to be explained is greater than the fraction specified
                      by ``pcf_variance``.
    Returns:
        List of linear dimensions as a fractions of the embeddings dimension.

    Let :math:`X` be a set of :math:`d` dimensional embeddings.
    Let :math:`\\lambda_1, \\ldots, \\lambda_d\\in\\mathbb{R}` be a set of eigenvalues
    of the covariance matrix of :math:`X` sorted in descending order.
    Then for a given value of desired explained variance :math:`r`,
    the number of principal components that explaines :math:`r\\cdot 100\\%%` variance is the largest integer
    :math:`n` such that

    .. math::
        \\frac{\\sum\\limits_{i = 1}^{n - 1}\\lambda_i}{\\sum\\limits_{i = 1}^{d}\\lambda_i} \\leq r

    The function returns

    .. math::
        \\frac{n}{d}

    See:

        `Principal Components Analysis`_

    .. _`Principal Components Analysis`:
        https://en.wikipedia.org/wiki/Principal_component_analysis

    Example:
        In the example bellow there are 4 vectors of length 10, and only first 4 dimensions have non-zero values.
        Its covariance matrix will have only 4 eigenvalues that are greater than 0, i.e. there are only 4 principal
        axes. So, in order to keep at least 50% of the information from the set, we need to keep 2 principal
        axes, and in order to keep all the information we need to keep 5 principal axes (one additional axis appears
        because the number of principal axes is superior to the desired explained variance threshold).

        >>> embeddings = torch.eye(4, 10, dtype=torch.float)
        >>> calc_pcf(embeddings, pcf_variance=(0.5, 1))
        tensor([0.2000, 0.5000])

    """
    # The code below mirrors code from scikit-learn repository:
    # https://github.com/scikit-learn/scikit-learn/blob/f3f51f9b6/sklearn/decomposition/_pca.py#L491
    _check_if_in_range(pcf_variance, 0, 1, "pcf_variance")
    try:
        pca = PCA(embeddings)
        n_components = pca.calc_principal_axes_number(pcf_variance).to(embeddings)
        metric = n_components / embeddings.shape[1]
    except Exception:
        # Mostly we handle the following error here:
        # >>> The algorithm failed to converge because the input matrix is ill-conditioned
        # >>> or has too many repeated singular values
        metric = [torch.tensor(float("nan"))] * len(pcf_variance)

    return metric


def _check_if_in_range(vals: Sequence[float], min_: float, max_: float, name: str) -> None:
    """
    Check whether the ``vals`` are in the range ``[min_, max_]``. Throw the ValueError if not.

    Args:
        vals: Sequence to check.
        min_: Minimal value of the range.
        max_: Maximal value of the range.
        name: Name of the variable to throw the ValueError for.
    """
    if len(vals) == 0:
        raise ValueError(f"{name} is expected to be not empty, but got {vals}")
    if not all(min_ <= x <= max_ for x in vals):
        raise ValueError(f"{name} is expected to contain numbers in range [{min_}, {max_}], but got {vals}")


__all__ = [
    "TMetricsDict",
    "calc_retrieval_metrics",
    "calc_topological_metrics",
    "reduce_metrics",
    "take_unreduced_metrics_by_mask",
    "calc_fnmr_at_fmr",
]



import warnings
from collections import defaultdict
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd

from oml.const import (
    BBOXES_COLUMNS,
    CATEGORIES_COLUMN,
    IS_GALLERY_COLUMN,
    IS_QUERY_COLUMN,
    LABELS_COLUMN,
    OBLIGATORY_COLUMNS,
    PATHS_COLUMN,
    SPLIT_COLUMN,
    TEXTS_COLUMN,
    X1_COLUMN,
    X2_COLUMN,
    Y1_COLUMN,
    Y2_COLUMN,
)


def check_retrieval_dataframe_format(
    df: Union[Path, str, pd.DataFrame], dataset_root: Optional[Path] = None, sep: str = ",", verbose: bool = True
) -> None:
    """
    Function checks if the dataset is in the correct format.

    Args:
        df: Path to ``.csv`` file or pandas DataFrame
        dataset_root: Path to the dataset root, set ``None`` if you used absolute paths in your dataframe
        sep: Separator used in ``.csv``
        verbose: Set ``True`` if you want to see warnings

    """
    if isinstance(df, (Path, str)):
        df = pd.read_csv(df, sep=sep, index_col=None)

    assert all(x in df.columns for x in OBLIGATORY_COLUMNS), df.columns
    assert (PATHS_COLUMN in df.columns) or (TEXTS_COLUMN in df.columns)

    assert set(df[SPLIT_COLUMN]).issubset({"train", "validation"}), set(df[SPLIT_COLUMN])

    mask_train = df[SPLIT_COLUMN] == "train"

    if mask_train.sum() > 0:
        q_train_vals = df[IS_QUERY_COLUMN][mask_train].unique()
        assert pd.isna(q_train_vals[0]) and len(q_train_vals) == 1, q_train_vals
        g_train_vals = df[IS_GALLERY_COLUMN][mask_train].unique()
        assert pd.isna(g_train_vals[0]) and len(g_train_vals) == 1, g_train_vals

    val_mask = ~mask_train

    if val_mask.sum() > 0:
        for split_field in [IS_QUERY_COLUMN, IS_GALLERY_COLUMN]:
            unq_values = set(df[split_field][val_mask])
            assert unq_values in [{False}, {True}, {False, True}], unq_values
        assert all(
            ((df[IS_QUERY_COLUMN][val_mask].astype(bool)) | df[IS_GALLERY_COLUMN][val_mask].astype(bool)).to_list()
        )

    assert np.issubdtype(df[LABELS_COLUMN].dtypes, np.integer)

    # we explicitly put ==True here because of Nones
    labels_query = set(df[LABELS_COLUMN][df[IS_QUERY_COLUMN] == True])  # noqa
    labels_gallery = set(df[LABELS_COLUMN][df[IS_GALLERY_COLUMN] == True])  # noqa
    assert labels_query.intersection(labels_gallery) == labels_query

    if PATHS_COLUMN in df.columns:
        if dataset_root is None:
            dataset_root = Path("")

        all_paths = df[PATHS_COLUMN].apply(lambda x: (dataset_root / x)).to_list()
        non_existing_paths = list(filter(lambda x: not x.exists(), all_paths))
        assert not non_existing_paths, f"Following paths do not exist: {non_existing_paths}"

    # check bboxes if exist
    if set(BBOXES_COLUMNS).intersection(set(list(df.columns))):
        assert all(x in df.columns for x in BBOXES_COLUMNS), df.columns

        bboxes_columns = df[BBOXES_COLUMNS]

        # here we check that for one example bounding box consists of four None (no bounding box) or have 4
        # integers as corners (checking that we don't use float indexes for the array)
        assert np.all(
            np.logical_or(
                np.isnan(bboxes_columns.values).sum(axis=1) == 4,
                (np.mod(bboxes_columns.values, 1) == 0).sum(axis=1) == 4,
            )
        )

        bboxes_df = df[~(df[X1_COLUMN].isna())]
        mask_good_x1_x2 = bboxes_df[X1_COLUMN] < bboxes_df[X2_COLUMN]
        mask_good_y1_y2 = bboxes_df[Y1_COLUMN] < bboxes_df[Y2_COLUMN]
        n_bad_x1_x2 = (~mask_good_x1_x2).sum()
        n_bad_y1_y2 = (~mask_good_y1_y2).sum()
        assert not n_bad_x1_x2, f"Number of bad x1/x2 pairs {n_bad_x1_x2}"
        assert not n_bad_y1_y2, f"Number of bad y1/y2 pairs {n_bad_y1_y2}"
        for coord in BBOXES_COLUMNS:
            assert all((bboxes_df[coord] >= 0).to_list()), coord

    if TEXTS_COLUMN in df.columns:
        assert isinstance(df.iloc[0][TEXTS_COLUMN], str)
        assert (df[TEXTS_COLUMN].apply(len) > 0).all()

    if CATEGORIES_COLUMN in df.columns:
        label_to_category = defaultdict(set)
        for _, row in df.iterrows():
            label_to_category[row[LABELS_COLUMN]].add(row[CATEGORIES_COLUMN])

        bad_categories = {k: v for k, v in label_to_category.items() if len(v) > 1}

        if bad_categories and verbose:
            warnings.warn(
                f"Note! You mapping between categories and labels is not bijection!"
                f"During the training and validation we will force it to be bijection by picking"
                f"one random category for each label."
                f"\n"
                f"{bad_categories}"
            )


__all__ = ["check_retrieval_dataframe_format"]



from typing import Any, List, Optional, Tuple, Union

import numpy as np
import torch
from torch import Tensor

from oml.interfaces.miners import ITripletsMinerInBatch, TTripletsIds
from oml.utils.misc_torch import pairwise_dist, take_2d


class NHardTripletsMiner(ITripletsMinerInBatch):
    """
    This miner selects hard triplets based on distances between features:

    - hard `positive` samples have large distance to the anchor sample

    - hard `negative` samples have small distance to the anchor sample

    Toward the end of the training, annotation errors can affect final metric. If you are not sure about the quality of
    your dataset, you can use range instead of integer value for parameters and exclude combinations with the largest
    distances. For example instead picking 5 positive examples, you can use examples from the 2nd hardest
    to the 5th one.

    """

    def __init__(
        self,
        n_positive: Union[Tuple[int, int], List[int], int] = 1,
        n_negative: Union[Tuple[int, int], List[int], int] = 1,
    ):
        """
        Args:
            n_positive: keep ``n_positive`` positive samples with large distances. If the value is a range, minimal
                value has to be less than the available amount of labels in batches
            n_negative: keep ``n_negative`` negative pipelines with small distances

        Note:
            If both parameters are 1, the miner is equivalent to ``HardTripletsMiner``.
            If both parameters are large enough, the miner can be equivalent to ``AllTripletsMiner``

        """

        self.positive_slice = slice(*self._parse_input_arg(n_positive))
        self.negative_slice = slice(*self._parse_input_arg(n_negative))

    @staticmethod
    def _parse_input_arg(n: Union[Tuple[int, int], List[int], int]) -> List[int]:
        if isinstance(n, int):
            n = [0, n]
        elif isinstance(n, (list, tuple)):
            n = list(n)
            n[0] -= 1
        else:
            raise TypeError("Unsupported type of argument. Must be int, tuple or list")

        assert n[1] > n[0]
        assert n[0] >= 0

        return n

    def _sample(
        self,
        features: Tensor,
        labels: List[int],
        *_: Any,
        ignore_anchor_mask: Optional[Union[List[int], Tensor, np.ndarray]] = None,
    ) -> TTripletsIds:
        """
        This method samples the hard triplets inside the batch based on distance between features.

        Args:
            features: Features with the shape of ``[batch_size, feature_size]``
            labels: Labels with the size of ``batch_size``
            *_: Anything here will be ignored
            ignore_anchor_mask: Parameter allows you to specify the ids of features that cannot be used as anchors. Can
            be useful with memory banks to create triplets in which at least one vector will have gradients.

        Returns:
            The batch of the triplets in the order below:
            ``(anchor, positive, negative)``
        """
        assert features.shape[0] == len(labels)

        dist_mat = pairwise_dist(x1=features, x2=features, p=2)

        ids_anchor, ids_pos, ids_neg = self._sample_from_distmat(
            distmat=dist_mat, labels=labels, ignore_anchor_mask=ignore_anchor_mask
        )

        return ids_anchor, ids_pos, ids_neg

    def _sample_from_distmat(
        self,
        distmat: Tensor,
        labels: List[int],
        *_: Any,
        ignore_anchor_mask: Optional[Union[List[int], Tensor, np.ndarray]] = None,
    ) -> TTripletsIds:
        if isinstance(labels, (list, tuple)):
            labels = torch.tensor(labels, device=distmat.device).long()

        if ignore_anchor_mask is None:
            ignore_anchor_mask = torch.zeros(len(distmat), dtype=torch.bool, device=distmat.device)

        all_ids_reduced = torch.arange(len(labels), device=distmat.device)[torch.logical_not(ignore_anchor_mask)]

        distmat_reduced = distmat[torch.logical_not(ignore_anchor_mask)]

        ids_highest_distance = torch.argsort(distmat_reduced, descending=True, dim=1)
        mask_same_label = labels[:, None] == labels[None, :]  # type: ignore
        mask_same_label.fill_diagonal_(False)
        mask_same_label = mask_same_label[torch.logical_not(ignore_anchor_mask)]
        mask_same_label_sorted_by_dist = take_2d(mask_same_label, ids_highest_distance)
        idx_anch_pos_reduced, idx_pos_sorted_by_dist = torch.nonzero(mask_same_label_sorted_by_dist, as_tuple=True)
        hardest_positive = ids_highest_distance[idx_anch_pos_reduced, idx_pos_sorted_by_dist]
        idx_anch_pos = all_ids_reduced[idx_anch_pos_reduced]

        ids_smallest_distance = torch.flip(ids_highest_distance, dims=(1,))
        mask_diff_label = labels[:, None] != labels[None, :]  # type: ignore
        mask_diff_label = mask_diff_label[torch.logical_not(ignore_anchor_mask)]
        diff_label_sorted_by_dist = take_2d(mask_diff_label, ids_smallest_distance)
        idx_anch_neg_reduced, idx_neg_sorted_by_dist = torch.nonzero(diff_label_sorted_by_dist, as_tuple=True)
        hardest_negative = ids_smallest_distance[idx_anch_neg_reduced, idx_neg_sorted_by_dist]
        idx_anch_neg = all_ids_reduced[idx_anch_neg_reduced]

        ids_a = []
        ids_p = []
        ids_n = []

        for idx_anch in torch.arange(len(labels), device=distmat.device)[torch.logical_not(ignore_anchor_mask)]:
            positives = hardest_positive[idx_anch_pos == idx_anch][self.positive_slice]
            negatives = hardest_negative[idx_anch_neg == idx_anch][self.negative_slice]

            i_pos, i_neg = list(zip(*torch.cartesian_prod(positives, negatives).tolist()))
            ids_a.extend([idx_anch.item()] * len(i_pos))
            ids_p.extend(i_pos)
            ids_n.extend(i_neg)

        return ids_a, ids_p, ids_n


__all__ = ["NHardTripletsMiner"]



from typing import Optional

import torch
import torch.nn.functional as F


@torch.no_grad()
def label_smoothing(
    y: torch.Tensor,
    num_classes: int,
    epsilon: float = 0.2,
    categories: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    """
    This function is doing `label smoothing <https://arxiv.org/pdf/1512.00567v3.pdf>`_.
    You can also use modified version, where the label is smoothed only for the category corresponding to sample's
    ground truth label. To use this, you should provide the ``categories`` argument: vector, for which i-th entry
    is a corresponding category for label ``i``.

    Args:
        y: Ground truth labels with the size of batch_size where each element is from 0 (inclusive) to
            num_classes (exclusive).
        num_classes: Number of classes in total
        epsilon: Power of smoothing. The biggest value in OHE-vector will be
            ``1 - epsilon + 1 / num_classes`` after the transformation
        categories: Vector for which i-th entry is a corresponding category for label ``i``. Optional, used for
            category-based label smoothing. In that case the biggest value in OHE-vector will be
            ``1 - epsilon + 1 / num_classes_of_the_same_category``, labels outside of the category will not change
    """
    assert epsilon < 1, "`epsilon` must be less than 1."
    ohe = F.one_hot(y, num_classes).float()
    if categories is not None:
        ohe *= 1 - epsilon
        same_category_mask = categories[y].tile(num_classes, 1).t() == categories
        return torch.where(same_category_mask, epsilon / same_category_mask.sum(-1).view(-1, 1), 0) + ohe
    else:
        ohe *= 1 - epsilon
        ohe += epsilon / num_classes
        return ohe



from collections import Counter
from typing import List

import numpy as np
import torch
from torch import Tensor

from oml.interfaces.miners import ITripletsMiner, TLabels, TTriplets, labels2list
from oml.utils.misc import find_value_ids
from oml.utils.misc_torch import pairwise_dist


class HardClusterMiner(ITripletsMiner):
    """
    This miner selects the hardest triplets based on the distance to mean vectors:
    anchor is a mean vector of features of i-th label in the batch, the hardest positive sample
    is the most distant from the anchor sample of anchor's label, the hardest negative
    sample is the closest mean vector of other labels.

    The batch must contain ``n_instances`` for ``n_labels`` where both values higher than 1.

    """

    def _check_input_labels(self, labels: List[int]) -> None:
        """
        Check if the labels list is valid: contains ``n_instances`` for each of ``n_labels``.

        Args:
            labels: Labels with the size of ``batch_size``

        Raises:
            ValueError: If the batch is invalid (contains different samples
                for labels, contains only one label, or only one sample for each label)

        """
        labels_counter = Counter(labels)
        n_instances = labels_counter[labels[0]]
        if not all(n == n_instances for n in labels_counter.values()):
            raise ValueError("Expected equal number of samples for each label")
        if len(labels_counter) <= 1:
            raise ValueError("Expected at least 2 labels in the batch")
        if n_instances == 1:
            raise ValueError("Expected more than one sample for each label")

    @staticmethod
    def _get_labels_mask(labels: List[int]) -> Tensor:
        """
        Generate matrix with the shape of ``[n_unique_labels, batch_size]``,
        where ``n_unique_labels`` is a number of unique labels
        in the batch; ``matrix[i, j]`` is ``True`` if j-th element of
        the batch relates to i-th label and ``False`` otherwise.

        Args:
            labels: Labels with the size of ``batch_size``

        Returns:
            Matrix of indices of labels in batch

        """
        unique_labels = sorted(np.unique(labels))
        labels_number = len(unique_labels)
        labels_mask = torch.zeros(size=(labels_number, len(labels)))
        for label_idx, label in enumerate(unique_labels):
            label_indices = find_value_ids(labels, label)
            labels_mask[label_idx][label_indices] = 1
        return labels_mask.type(torch.bool)

    @staticmethod
    def _count_intra_label_distances(embeddings: Tensor, mean_vectors: Tensor) -> Tensor:
        """
        Count matrix of distances from mean vector of each label to it's
        samples embeddings.

        Args:
            embeddings: Tensor with the shape of ``[n_labels, n_instances, embed_dim]``
            mean_vectors: Tensor with the shape of ``[n_labels, embed_dim]`` -- mean vectors
                of each label in the batch

        Returns:
            Tensor with the shape of ``[n_labels, n_instances]`` -- matrix of distances from mean vectors to
                related samples in the batch

        """
        n_labels, n_instances, embed_dim = embeddings.shape
        # Create (n_labels, n_instances, embed_dim) tensor of mean vectors for each label
        mean_vectors = mean_vectors.unsqueeze(1).repeat((1, n_instances, 1))
        # Count euclidean distance between embeddings and mean vectors
        distances = torch.pow(embeddings - mean_vectors, 2).sum(2)
        return distances

    @staticmethod
    def _count_inter_label_distances(mean_vectors: Tensor) -> Tensor:
        """
        Count matrix of distances from mean vectors of labels to each other

        Args:
            mean_vectors: Tensor with the shape of ``[n_labels, embed_dim]`` -- mean vectors
                of the labels

        Returns:
            Tensor with the shape of ``[n_labels, n_labels]`` -- matrix of distances between mean vectors

        """
        distance = pairwise_dist(x1=mean_vectors, x2=mean_vectors, p=2)
        return distance

    @staticmethod
    def _fill_diagonal(matrix: Tensor, value: float) -> Tensor:
        """
        Set diagonal elements with the value.

        Args:
            matrix: Tensor with the shape of ``[n_labels, n_labels]``
            value: Value that diagonal should be filled with

        Returns:
            Modified matrix with inf on diagonal

        """
        n_labels, _ = matrix.shape
        indices = torch.diag(torch.ones(n_labels)).type(torch.bool)
        matrix[indices] = value
        return matrix

    def sample(self, features: Tensor, labels: TLabels) -> TTriplets:
        """
        This method samples the hardest triplets in the batch.

        Args:
            features: Tensor with the shape of ``[batch_size, embed_dim]`` that contains
                ``n_instances`` for each of ``n_labels``
            labels: Labels with the size of ``batch_size``

        Returns:
            ``n_labels`` triplets in the form of ``(mean_vector, positive, negative_mean_vector)``

        """
        # Convert labels to list
        labels = labels2list(labels)
        self._check_input_labels(labels)

        # Get matrix of indices of labels in batch
        labels_mask = self._get_labels_mask(labels)
        n_labels = labels_mask.shape[0]

        embed_dim = features.shape[-1]
        # Reshape embeddings to groups of (n_labels, n_instances, embed_dim) ones,
        # each i-th group contains embeddings of i-th label.
        features = features.repeat((n_labels, 1, 1))
        features = features[labels_mask].view((n_labels, -1, embed_dim))

        # Count mean vectors for each label in batch
        mean_vectors = features.mean(1)

        d_intra = self._count_intra_label_distances(features, mean_vectors)
        # Count the distances to the sample farthest from mean vector
        # for each label.
        pos_indices = d_intra.max(1).indices
        # Count matrix of distances from mean vectors to each other
        d_inter = self._count_inter_label_distances(mean_vectors)
        # For each label mean vector get the closest mean vector
        d_inter = self._fill_diagonal(d_inter, float("inf"))
        neg_indices = d_inter.min(1).indices
        positives = torch.stack([features[idx][pos_idx] for idx, pos_idx in enumerate(pos_indices)])

        return mean_vectors, positives, mean_vectors[neg_indices]


__all__ = ["HardClusterMiner"]



# type: ignore
# flake8: noqa

# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Mostly copy-paste from timm library.
https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py
"""
import math
import warnings
from functools import partial

import torch
import torch.nn as nn
from torch import Tensor


def trunc_normal_(tensor, mean=0.0, std=1.0, a=-2.0, b=2.0):
    # type: (Tensor, float, float, float, float) -> Tensor
    return _no_grad_trunc_normal_(tensor, mean, std, a, b)


def _no_grad_trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1.0 + math.erf(x / math.sqrt(2.0))) / 2.0

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn(
            "mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
            "The distribution of values may be incorrect.",
            stacklevel=2,
        )

    with torch.no_grad():
        # Values are generated by using a truncated uniform distribution and
        # then using the inverse CDF for the normal distribution.
        # Get upper and lower cdf values
        l = norm_cdf((a - mean) / std)
        u = norm_cdf((b - mean) / std)

        # Uniformly fill tensor with values from [l, u], then translate to
        # [2l-1, 2u-1].
        tensor.uniform_(2 * l - 1, 2 * u - 1)

        # Use inverse cdf transform for normal distribution to get truncated
        # standard normal
        tensor.erfinv_()

        # Transform to proper mean, std
        tensor.mul_(std * math.sqrt(2.0))
        tensor.add_(mean)

        # Clamp to ensure it's in the proper range
        tensor.clamp_(min=a, max=b)
        return tensor


def drop_path(x, drop_prob: float = 0.0, training: bool = False):
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)


class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.0):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0.0, proj_drop=0.0):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x, attn


class Block(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward(self, x, return_attention: bool = False):
        y, attn = self.attn(self.norm1(x))
        if return_attention:
            return attn
        x = x + self.drop_path(y)
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x


class PatchEmbed(nn.Module):
    """Image to Patch Embedding"""

    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        num_patches = (img_size // patch_size) * (img_size // patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)

    def forward(self, x):
        B, C, H, W = x.shape
        x = self.proj(x).flatten(2).transpose(1, 2)
        return x


class VisionTransformer(nn.Module):
    """Vision Transformer"""

    def __init__(
        self,
        img_size=[224],
        patch_size=16,
        in_chans=3,
        num_classes=0,
        embed_dim=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4.0,
        qkv_bias=False,
        qk_scale=None,
        drop_rate=0.0,
        attn_drop_rate=0.0,
        drop_path_rate=0.0,
        norm_layer=nn.LayerNorm,
        **kwargs
    ):
        super().__init__()
        self.num_features = self.embed_dim = embed_dim

        self.patch_embed = PatchEmbed(
            img_size=img_size[0], patch_size=patch_size, in_chans=in_chans, embed_dim=embed_dim
        )
        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))
        self.pos_drop = nn.Dropout(p=drop_rate)

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.blocks = nn.ModuleList(
            [
                Block(
                    dim=embed_dim,
                    num_heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    qkv_bias=qkv_bias,
                    qk_scale=qk_scale,
                    drop=drop_rate,
                    attn_drop=attn_drop_rate,
                    drop_path=dpr[i],
                    norm_layer=norm_layer,
                )
                for i in range(depth)
            ]
        )
        self.norm = norm_layer(embed_dim)

        # Classifier head
        self.head = nn.Linear(embed_dim, num_classes) if num_classes > 0 else nn.Identity()

        trunc_normal_(self.pos_embed, std=0.02)
        trunc_normal_(self.cls_token, std=0.02)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def interpolate_pos_encoding(self, x, w: int, h: int):
        npatch = x.shape[1] - 1
        N = self.pos_embed.shape[1] - 1
        if npatch == N and w == h:
            return self.pos_embed
        class_pos_embed = self.pos_embed[:, 0]
        patch_pos_embed = self.pos_embed[:, 1:]
        dim = x.shape[-1]
        w0 = w // self.patch_embed.patch_size
        h0 = h // self.patch_embed.patch_size
        # we add a small number to avoid floating point error in the interpolation
        # see discussion at https://github.com/facebookresearch/dino/issues/8
        w0, h0 = w0 + 0.1, h0 + 0.1
        patch_pos_embed = nn.functional.interpolate(
            patch_pos_embed.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(0, 3, 1, 2),
            scale_factor=(w0 / math.sqrt(N), h0 / math.sqrt(N)),
            mode="bicubic",
        )
        assert int(w0) == patch_pos_embed.shape[-2] and int(h0) == patch_pos_embed.shape[-1]
        patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
        return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1)

    def prepare_tokens(self, x):
        B, nc, w, h = x.shape
        x = self.patch_embed(x)  # patch linear embedding

        # add the [CLS] token to the embed patch tokens
        cls_tokens = self.cls_token.expand(B, -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)

        # add positional encoding to each token
        x = x + self.interpolate_pos_encoding(x, w, h)

        return self.pos_drop(x)

    def forward(self, x):
        x = self.prepare_tokens(x)
        for blk in self.blocks:
            x = blk(x)
        x = self.norm(x)
        return x[:, 0]

    def get_last_selfattention(self, x):
        x = self.prepare_tokens(x)
        for i, blk in enumerate(self.blocks):
            if i < len(self.blocks) - 1:
                x = blk(x)
            else:
                # return attention of the last block
                return blk(x, return_attention=True)

    def get_intermediate_layers(self, x, n=1):
        x = self.prepare_tokens(x)
        # we return the output tokens from the `n` last blocks
        output = []
        for i, blk in enumerate(self.blocks):
            x = blk(x)
            if len(self.blocks) - i <= n:
                output.append(self.norm(x))
        return output


def vit_tiny(patch_size=16, **kwargs):
    model = VisionTransformer(
        patch_size=patch_size,
        embed_dim=192,
        depth=12,
        num_heads=3,
        mlp_ratio=4,
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        **kwargs
    )
    return model


def vit_small(patch_size=16, **kwargs):
    model = VisionTransformer(
        patch_size=patch_size,
        embed_dim=384,
        depth=12,
        num_heads=6,
        mlp_ratio=4,
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        **kwargs
    )
    return model


def vit_base(patch_size=16, **kwargs):
    model = VisionTransformer(
        patch_size=patch_size,
        embed_dim=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4,
        qkv_bias=True,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        **kwargs
    )
    return model


class DINOHead(nn.Module):
    def __init__(
        self, in_dim, out_dim, use_bn=False, norm_last_layer=True, nlayers=3, hidden_dim=2048, bottleneck_dim=256
    ):
        super().__init__()
        nlayers = max(nlayers, 1)
        if nlayers == 1:
            self.mlp = nn.Linear(in_dim, bottleneck_dim)
        else:
            layers = [nn.Linear(in_dim, hidden_dim)]
            if use_bn:
                layers.append(nn.BatchNorm1d(hidden_dim))
            layers.append(nn.GELU())
            for _ in range(nlayers - 2):
                layers.append(nn.Linear(hidden_dim, hidden_dim))
                if use_bn:
                    layers.append(nn.BatchNorm1d(hidden_dim))
                layers.append(nn.GELU())
            layers.append(nn.Linear(hidden_dim, bottleneck_dim))
            self.mlp = nn.Sequential(*layers)
        self.apply(self._init_weights)
        self.last_layer = nn.utils.weight_norm(nn.Linear(bottleneck_dim, out_dim, bias=False))
        self.last_layer.weight_g.data.fill_(1)
        if norm_last_layer:
            self.last_layer.weight_g.requires_grad = False

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.mlp(x)
        x = nn.functional.normalize(x, dim=-1, p=2)
        x = self.last_layer(x)
        return x



import numpy as np
import torch
from torch import nn


class GEM(nn.Module):
    def __init__(self, p: float, eps: float = 1e-6):
        """
        Generalised Mean Pooling (GEM)
        https://paperswithcode.com/method/generalized-mean-pooling

        Args:
            p: if p == 1 it's average pooling, if p == inf it's max-pooling
            eps: eps for numerical stability
        """
        super(GEM, self).__init__()
        self.p = p
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = torch.clip(x, min=self.eps, max=np.inf)
        x = torch.pow(x, self.p)

        bs, feat_dim, h, w = x.shape
        x = x.view(bs, feat_dim, h * w)

        x = x.mean(axis=-1)
        x = torch.pow(x, (1.0 / self.p))
        return x


__all__ = ["GEM"]



# type: ignore
# flake8: noqa

import math
import warnings

import torch
import torch.nn as nn
from torch.utils.checkpoint import checkpoint
from torchvision.transforms import (
    CenterCrop,
    Compose,
    InterpolationMode,
    Normalize,
    Resize,
    ToTensor,
)

# ============== CODE FROM TIMM ==============
# https://github.com/huggingface/pytorch-image-models/blob/
# c241081251323dfc5e8dc799d49740c48cc9096f/timm/layers/weight_init.py#L8


def _trunc_normal_(tensor, mean, std, a, b):
    # Cut & paste from PyTorch official master until it's in a few official releases - RW
    # Method based on https://people.sc.fsu.edu/~jburkardt/presentations/truncated_normal.pdf
    def norm_cdf(x):
        # Computes standard normal cumulative distribution function
        return (1.0 + math.erf(x / math.sqrt(2.0))) / 2.0

    if (mean < a - 2 * std) or (mean > b + 2 * std):
        warnings.warn(
            "mean is more than 2 std from [a, b] in nn.init.trunc_normal_. "
            "The distribution of values may be incorrect.",
            stacklevel=2,
        )

    # Values are generated by using a truncated uniform distribution and
    # then using the inverse CDF for the normal distribution.
    # Get upper and lower cdf values
    l = norm_cdf((a - mean) / std)
    u = norm_cdf((b - mean) / std)

    # Uniformly fill tensor with values from [l, u], then translate to
    # [2l-1, 2u-1].
    tensor.uniform_(2 * l - 1, 2 * u - 1)

    # Use inverse cdf transform for normal distribution to get truncated
    # standard normal
    tensor.erfinv_()

    # Transform to proper mean, std
    tensor.mul_(std * math.sqrt(2.0))
    tensor.add_(mean)

    # Clamp to ensure it's in the proper range
    tensor.clamp_(min=a, max=b)
    return tensor


def trunc_normal_(tensor, mean=0.0, std=1.0, a=-2.0, b=2.0):
    r"""Fills the input Tensor with values drawn from a truncated
    normal distribution. The values are effectively drawn from the
    normal distribution :math:`\mathcal{N}(\text{mean}, \text{std}^2)`
    with values outside :math:`[a, b]` redrawn until they are within
    the bounds. The method used for generating the random values works
    best when :math:`a \leq \text{mean} \leq b`.

    NOTE: this impl is similar to the PyTorch trunc_normal_, the bounds [a, b] are
    applied while sampling the normal with mean/std applied, therefore a, b args
    should be adjusted to match the range of mean, std args.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        mean: the mean of the normal distribution
        std: the standard deviation of the normal distribution
        a: the minimum cutoff value
        b: the maximum cutoff value
    """
    with torch.no_grad():
        return _trunc_normal_(tensor, mean, std, a, b)


def drop_path(x, drop_prob: float = 0.0, training: bool = False, scale_by_keep: bool = True):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (x.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
    if keep_prob > 0.0 and scale_by_keep:
        random_tensor.div_(keep_prob)
    return x * random_tensor


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob: float = 0.0, scale_by_keep: bool = True):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training, self.scale_by_keep)

    def extra_repr(self):
        return f"drop_prob={round(self.drop_prob, 3):0.3f}"


# ============== CODE FROM UNICOM ==============
# https://github.com/deepglint/unicom/blob/main/unicom/vision_transformer.py


class VisionTransformer(nn.Module):
    def __init__(
        self,
        input_size=224,
        patch_size=32,
        in_channels=3,
        dim=768,
        embedding_size=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4,
        drop_path_rate=0.0,
        using_checkpoint=True,
    ):
        super().__init__()
        self.dim = dim
        self.patch_embed = PatchEmbedding(
            input_size,
            patch_size,
            in_channels,
            dim,
        )
        self.pos_embed = nn.Parameter(torch.zeros(1, self.patch_embed.num_patches, dim))
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]

        self.blocks = nn.ModuleList(
            [
                Block(dim, num_heads, mlp_ratio, dpr[i], self.patch_embed.num_patches, using_checkpoint)
                for i in range(depth)
            ]
        )
        self.norm = nn.LayerNorm(dim)

        self.feature = nn.Sequential(
            nn.Linear(dim * self.patch_embed.num_patches, dim, False),
            nn.BatchNorm1d(dim, eps=2e-5),
            nn.Linear(dim, embedding_size, False),
            nn.BatchNorm1d(embedding_size, eps=2e-5),
        )

        trunc_normal_(self.pos_embed, std=0.02)
        self.apply(self._init_weights)
        self.extra_gflops = 0.0
        for _block in self.blocks:
            self.extra_gflops += _block.extra_gflops

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def forward_features(self, x):
        B = x.shape[0]
        x = self.patch_embed(x)
        x = x + self.pos_embed
        for func in self.blocks:
            x = func(x)
        x = self.norm(x.float())
        return torch.reshape(x, (B, self.patch_embed.num_patches * self.dim))

    def forward(self, x):
        x = self.forward_features(x)
        x = self.feature(x)
        return x


class Mlp(nn.Module):
    def __init__(self, dim, dim_hidden):
        super().__init__()
        self.fc1 = nn.Linear(dim, dim_hidden)
        self.act = nn.ReLU6()
        self.fc2 = nn.Linear(dim_hidden, dim)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5
        self.qkv = nn.Linear(dim, dim * 3, bias=False)
        self.proj = nn.Linear(dim, dim)

    def forward(self, x):
        with torch.cuda.amp.autocast(True):
            B, L, D = x.shape
            qkv = self.qkv(x).reshape(B, L, 3, self.num_heads, D // self.num_heads).permute(2, 0, 3, 1, 4)
        with torch.cuda.amp.autocast(False):
            q, k, v = qkv[0].float(), qkv[1].float(), qkv[2].float()
            attn = (q @ k.transpose(-2, -1)) * self.scale
            attn = attn.softmax(dim=-1)
            x = (attn @ v).transpose(1, 2).reshape(B, L, D)
        with torch.cuda.amp.autocast(True):
            x = self.proj(x)
        return x


class Block(nn.Module):
    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: int = 4,
        drop_path: float = 0.0,
        patch_n: int = 32,
        using_checkpoint=False,
    ):
        super().__init__()
        self.using_checkpoint = using_checkpoint
        self.norm1 = nn.LayerNorm(dim)
        self.norm2 = nn.LayerNorm(dim)
        self.attn = Attention(dim, num_heads)
        if drop_path > 0:
            self.drop_path = DropPath(drop_path)
        else:
            self.drop_path = nn.Identity()
        self.mlp = Mlp(dim, dim * mlp_ratio)
        self.extra_gflops = (num_heads * patch_n * (dim // num_heads) * patch_n * 2) / (1000**3)

    def forward_impl(self, x):
        with torch.cuda.amp.autocast(True):
            x = x + self.drop_path(self.attn(self.norm1(x)))
            x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x

    def forward(self, x):
        if self.using_checkpoint:
            return checkpoint(self.forward_impl, x)
        else:
            return self.forward_impl(x)


class PatchEmbedding(nn.Module):
    def __init__(self, input_size=224, patch_size=32, in_channels: int = 3, dim: int = 768):
        super().__init__()
        if isinstance(input_size, int):
            input_size = (input_size, input_size)
        if isinstance(patch_size, int):
            patch_size = (patch_size, patch_size)
        H = input_size[0] // patch_size[0]
        W = input_size[1] // patch_size[1]
        self.num_patches = H * W
        self.proj = nn.Conv2d(in_channels, dim, kernel_size=patch_size, stride=patch_size)

    def forward(self, x):
        x = self.proj(x).flatten(2).transpose(1, 2)
        return x


def build_model(name="ViT-L/14@336px", using_checkpoint: bool = True):
    if name == "ViT-B/32":
        model = VisionTransformer(
            input_size=224,
            patch_size=32,
            in_channels=3,
            dim=768,
            embedding_size=512,
            depth=12,
            num_heads=12,
            drop_path_rate=0.1,
            using_checkpoint=using_checkpoint,
        )
    elif name == "ViT-B/16":
        model = VisionTransformer(
            input_size=224,
            patch_size=16,
            in_channels=3,
            dim=768,
            embedding_size=768,
            depth=12,
            num_heads=12,
            drop_path_rate=0.1,
            using_checkpoint=using_checkpoint,
        )
    elif name == "ViT-L/14":
        model = VisionTransformer(
            input_size=224,
            patch_size=14,
            in_channels=3,
            dim=1024,
            embedding_size=768,
            depth=24,
            num_heads=16,
            drop_path_rate=0.1,
            using_checkpoint=using_checkpoint,
        )
    elif name == "ViT-L/14@336px":
        model = VisionTransformer(
            input_size=336,
            patch_size=14,
            in_channels=3,
            dim=1024,
            embedding_size=768,
            depth=24,
            num_heads=16,
            drop_path_rate=0.1,
            using_checkpoint=using_checkpoint,
        )
    return model


def _convert_image_to_rgb(image):
    return image.convert("RGB")


def transform(im_size):
    return Compose(
        [
            Resize(im_size, interpolation=InterpolationMode.BICUBIC),
            CenterCrop(im_size),
            _convert_image_to_rgb,
            ToTensor(),
            Normalize((0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711)),
        ]
    )


def load_model_and_transform(name="ViT-L/14@336px"):
    if name == "ViT-B/32":
        return build_model(name), transform(224)
    elif name == "ViT-B/16":
        return build_model(name), transform(224)
    elif name == "ViT-L/14":
        return build_model(name), transform(224)
    elif name == "ViT-L/14@336px":
        return build_model(name), transform(336)
    else:
        raise



from pathlib import Path
from typing import Any, Dict, Optional, Union

import numpy as np
import PIL.Image
import torch
from PIL.Image import Image as TPILImage
from torch import nn
from torchvision.models import resnet18, resnet34, resnet50, resnet101, resnet152

from oml.interfaces.models import IExtractor
from oml.models.resnet.pooling import GEM
from oml.models.utils import (
    remove_criterion_in_state_dict,
    remove_prefix_from_state_dict,
)
from oml.transforms.images.albumentations import get_normalisation_albu
from oml.utils.io import download_checkpoint
from oml.utils.misc_torch import get_device, normalise


def resnet50_projector() -> nn.Module:
    model = resnet50(weights=None, num_classes=128)
    model.fc = nn.Sequential(
        nn.Linear(model.fc.weight.shape[1], model.fc.weight.shape[1]),
        nn.ReLU(),
        model.fc,
    )
    return model


class ResnetExtractor(IExtractor):
    """
    The base class for the extractors that follow ResNet architecture.

    """

    constructors = {
        "resnet18": resnet18,
        "resnet34": resnet34,
        "resnet50": resnet50,
        "resnet50_projector": resnet50_projector,
        "resnet101": resnet101,
        "resnet152": resnet152,
    }

    pretrained_models = {
        "resnet50_moco_v2": {
            "url": "https://dl.fbaipublicfiles.com/moco/moco_checkpoints/moco_v2_800ep/moco_v2_800ep_pretrain.pth.tar",
            "hash": "a04e12f8",
            "fname": "moco_v2_800ep_pretrain.pth.tar",
            "init_args": {"arch": "resnet50_projector", "remove_fc": True, "normalise_features": False, "gem_p": 5.0},
        },
        "resnet18_imagenet1k_v1": {
            "init_args": {"arch": "resnet18", "remove_fc": True, "normalise_features": False, "gem_p": None}
        },
        "resnet34_imagenet1k_v1": {
            "init_args": {"arch": "resnet34", "remove_fc": True, "normalise_features": False, "gem_p": None}
        },
        "resnet50_imagenet1k_v1": {
            "init_args": {"arch": "resnet50", "remove_fc": True, "normalise_features": False, "gem_p": None},
        },
        "resnet101_imagenet1k_v1": {
            "init_args": {"arch": "resnet101", "remove_fc": True, "normalise_features": False, "gem_p": None},
        },
        "resnet152_imagenet1k_v1": {
            "init_args": {"arch": "resnet152", "remove_fc": True, "normalise_features": False, "gem_p": None},
        },
    }

    def __init__(
        self,
        weights: Optional[Union[Path, str]],
        arch: str,
        gem_p: Optional[float],
        remove_fc: bool,
        normalise_features: bool,
    ):
        """

        Args:
            weights: Path to weights or a special key to download pretrained checkpoint, use ``None`` to randomly
             initialize model's weights. You can check the available pretrained checkpoints
             in ``self.pretrained_models``.
            arch: Different types of ResNet, please, check ``self.constructors``
            gem_p: Value of power in `Generalized Mean Pooling` that we use as the replacement for the default one
             (if ``gem_p == 1`` or ``None`` it's just a normal average pooling and if ``gem_p -> inf`` it's max-pooling)
            remove_fc: Set ``True`` if you want to remove the last fully connected layer. Note, that having this layer
              is obligatory for calling ``draw_gradcam()`` method
            normalise_features: Set ``True`` to normalise output features

        """
        assert arch in self.constructors.keys()
        super(ResnetExtractor, self).__init__()

        self.arch = arch
        self.gem_p = gem_p
        self.remove_fc = remove_fc
        self.normalise_features = normalise_features

        constructor = self.constructors[self.arch]
        self.model = constructor()

        if gem_p is not None:
            self.model.avgpool = GEM(p=gem_p)

        if weights is None:
            if self.remove_fc:
                self.model.fc = nn.Identity()
            return

        elif weights == "resnet50_moco_v2":
            pretrained = self.pretrained_models[weights]  # type: ignore
            moco_path = download_checkpoint(
                url_or_fid=pretrained["url"],  # type: ignore
                hash_md5=pretrained["hash"],  # type: ignore
                fname=pretrained["fname"],  # type: ignore
            )
            state_dict = load_moco_weights(moco_path)

        elif str(weights).endswith("_imagenet1k_v1"):
            state_dict = constructor(weights="IMAGENET1K_V1").state_dict()

        else:
            state_dict = torch.load(weights, map_location="cpu")

        state_dict = state_dict["state_dict"] if "state_dict" in state_dict.keys() else state_dict
        state_dict = remove_criterion_in_state_dict(state_dict)  # type: ignore
        state_dict = remove_prefix_from_state_dict(state_dict, "layer4.")  # type: ignore

        if self.remove_fc:
            state_dict.pop("fc.weight", None)
            state_dict.pop("fc.bias", None)
            if arch != "resnet50_projector":
                self.model.fc = nn.Identity()

        self.model.load_state_dict(state_dict, strict=True)

        if self.remove_fc:
            self.model.fc = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.model(x)

        if self.normalise_features:
            x = normalise(x)

        return x

    def get_last_conv_channels(self) -> int:
        last_block = self.model.layer4[-1]
        if self.arch in ("resnet18", "resnet34"):
            n_out_channels = last_block.conv2.out_channels
        else:
            # resnet50, resnet101, resnet152
            n_out_channels = last_block.conv3.out_channels
        return n_out_channels

    @property
    def feat_dim(self) -> int:
        if isinstance(self.model.fc, torch.nn.Identity):
            return self.get_last_conv_channels()
        elif isinstance(self.model.fc, torch.nn.Linear):
            return self.model.fc.out_features
        else:
            # 2-layer mlp case
            return self.model.fc[-1].out_features

    def draw_gradcam(self, image: Union[np.ndarray, TPILImage]) -> Union[np.ndarray, TPILImage]:
        """
        Args:
            image: An image with pixel values in the range of ``[0..255]``.

        Returns:
            An image with drawn gradients.

        Visualization of the gradients on a particular image using `GradCam`_.

        .. _GradCam: https://arxiv.org/abs/1610.02391

        """
        # this is the optional dependency
        from pytorch_grad_cam import GradCAM
        from pytorch_grad_cam.utils.image import show_cam_on_image

        if self.remove_fc:
            raise ValueError("This method does not work if there is no FC layer in the model.")

        need_to_convert = not isinstance(image, np.ndarray)

        if need_to_convert:
            image = np.asarray(image)

        device = get_device(self.model)
        image_tensor = get_normalisation_albu()(image=image)["image"].to(device)
        cam = GradCAM(model=self.model, target_layers=[self.model.layer4[-1]], use_cuda=device != "cpu")
        gray_image = cam(image_tensor.unsqueeze(0), None)[0]
        img_with_grads = show_cam_on_image(image / 255, gray_image)

        if need_to_convert:
            img_with_grads = PIL.Image.fromarray(img_with_grads)

        return img_with_grads


def load_moco_weights(path_to_model: Union[str, Path]) -> Dict[str, Any]:
    """
    Args:
        path_to_model: Path to model trained using original
           code from MoCo repository:
           https://github.com/facebookresearch/moco

    Returns:
        State dict without weights of student

    """
    checkpoint = torch.load(path_to_model, map_location="cpu")

    state_dict = checkpoint["state_dict"]
    for key in list(state_dict.keys()):
        # retain only encoder_q up to before the embedding layer
        if key.startswith("module.encoder_q"):
            new_key = key[len("module.encoder_q.") :]
            state_dict[new_key] = state_dict[key]
        del state_dict[key]

    return state_dict


__all__ = ["ResnetExtractor"]



from pathlib import Path
from typing import Any, Dict, List, Tuple

import albumentations as albu

import oml.models.vit_unicom.external.vision_transformer as unicom  # type: ignore
from oml.transforms.images.albumentations import (
    get_augs_albu,
    get_normalisation_albu,
    get_normalisation_resize_albu,
    get_normalisation_resize_albu_clip,
)
from oml.transforms.images.torchvision import (
    get_augs_hypvit,
    get_augs_torch,
    get_normalisation_resize_hypvit,
    get_normalisation_resize_torch,
    get_normalisation_torch,
)
from oml.transforms.images.utils import TTransforms, get_im_reader_for_transforms
from oml.utils.images.images import TImReader
from oml.utils.misc import TCfg, dictconfig_to_dict

TRANSFORMS_ALBU = {
    "augs_albu": get_augs_albu,
    "norm_albu": get_normalisation_albu,
    "norm_resize_albu": get_normalisation_resize_albu,
    "norm_resize_albu_clip": get_normalisation_resize_albu_clip,
}

TRANSFORMS_TORCH = {
    "augs_torch": get_augs_torch,
    "norm_torch": get_normalisation_torch,
    "norm_resize_torch": get_normalisation_resize_torch,
    "augs_hypvit_torch": get_augs_hypvit,
    "norm_resize_hypvit_torch": get_normalisation_resize_hypvit,
    "unicom_transforms": unicom.transform,  # type: ignore
}

TRANSFORMS_REGISTRY = {**TRANSFORMS_ALBU, **TRANSFORMS_TORCH}


def get_transforms(name: str, **kwargs: Dict[str, Any]) -> TTransforms:
    augs = TRANSFORMS_REGISTRY[name](**kwargs)  # type: ignore
    return augs


def get_transforms_by_cfg(cfg: TCfg) -> TTransforms:
    cfg = dictconfig_to_dict(cfg)
    return get_transforms(name=cfg["name"], **cfg["args"])


TRANSFORMS_FOR_PRETRAINED = {
    "resnet50_moco_v2": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "resnet18_imagenet1k_v1": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "resnet34_imagenet1k_v1": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "resnet50_imagenet1k_v1": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "resnet101_imagenet1k_v1": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "resnet152_imagenet1k_v1": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "vitb8_dino": get_normalisation_resize_torch(im_size=224),
    "vitb16_dino": get_normalisation_resize_torch(im_size=224),
    "vits8_dino": get_normalisation_resize_torch(im_size=224),
    "vits16_dino": get_normalisation_resize_torch(im_size=224),
    "vitb14_dinov2": get_normalisation_resize_hypvit(im_size=224),
    "vitb14_reg_dinov2": get_normalisation_resize_hypvit(im_size=224),
    "vitl14_dinov2": get_normalisation_resize_hypvit(im_size=224),
    "vitl14_reg_dinov2": get_normalisation_resize_hypvit(im_size=224),
    "vits14_dinov2": get_normalisation_resize_hypvit(im_size=224),
    "vits14_reg_dinov2": get_normalisation_resize_hypvit(im_size=224),
    "sber_vitb32_224": get_normalisation_resize_albu_clip(im_size=224),
    "sber_vitb16_224": get_normalisation_resize_albu_clip(im_size=224),
    "sber_vitl14_224": get_normalisation_resize_albu_clip(im_size=224),
    "openai_vitb32_224": get_normalisation_resize_albu_clip(im_size=224),
    "openai_vitb16_224": get_normalisation_resize_albu_clip(im_size=224),
    "openai_vitl14_224": get_normalisation_resize_albu_clip(im_size=224),
    "vits16_inshop": get_normalisation_resize_hypvit(im_size=224, crop_size=224),
    "vits16_sop": get_normalisation_resize_hypvit(im_size=224, crop_size=224),
    "vits16_cars": get_normalisation_resize_albu(im_size=224),
    "vits16_cub": get_normalisation_resize_albu(im_size=224),
    "vits16_224_mlp_384_inshop": get_normalisation_resize_hypvit(im_size=256, crop_size=224),
    "vitb32_unicom": unicom.transform(224),  # type: ignore
    "vitb16_unicom": unicom.transform(224),  # type: ignore
    "vitl14_unicom": unicom.transform(224),  # type: ignore
    "vitl14_336px_unicom": unicom.transform(336),  # type: ignore
}


def get_transforms_for_pretrained(weights: str) -> Tuple[TTransforms, TImReader]:
    transforms = TRANSFORMS_FOR_PRETRAINED[weights]
    im_reader = get_im_reader_for_transforms(transforms)
    return transforms, im_reader


def save_transforms_as_files(cfg: TCfg) -> List[Tuple[str, str]]:
    """
    Function saves transforms as files in local filesystem and returns list of tuples
    (transform_cfg_key, path_to_transform_file) for each transform
    """
    keys_files = []

    for key, val in cfg.items():
        if "transforms" in key:
            try:
                transforms = get_transforms_by_cfg(cfg[key])
                if isinstance(transforms, albu.Compose):
                    transforms_file = str(Path(".hydra/") / f"{key}.yaml") if Path(".hydra").exists() else f"{key}.yaml"
                    albu.save(filepath=transforms_file, transform=transforms, data_format="yaml")
                    keys_files.append((key, transforms_file))
            except Exception:
                print(f"We are not able to interpret {key} as albumentations transforms and log them as a file.")
    return keys_files


__all__ = [
    "TRANSFORMS_TORCH",
    "TRANSFORMS_ALBU",
    "TRANSFORMS_REGISTRY",
    "get_transforms",
    "get_transforms_by_cfg",
    "get_transforms_for_pretrained",
    "save_transforms_as_files",
]



import inspect
import logging
import warnings
from typing import Any, Dict, List, Sequence, Union

from torch.utils.data import (
    BatchSampler,
    DataLoader,
    Dataset,
    DistributedSampler,
    Sampler,
)

from oml.ddp.utils import WarningDDP, is_ddp
from oml.interfaces.samplers import IBatchSampler

TAllSamplers = Union[BatchSampler, Sampler, IBatchSampler]


class _Sampler2Dataset(Dataset):
    def __init__(self, sampler: TAllSamplers):
        # We read sampler in __getitem__ due to the seed between calling of __init__ and __getitem__ can be changed
        self.sampler_read = None
        self.sampler = sampler

    def __getitem__(self, item: int) -> Union[int, List[int]]:
        if self.sampler_read is None:
            self.sampler_read = list(self.sampler)  # type: ignore

        return self.sampler_read[item]  # type: ignore

    def __len__(self) -> int:
        return len(self.sampler)  # type: ignore


class DDPSamplerWrapper(DistributedSampler):
    """
    This is a wrapper to allow using custom sampler in DDP mode.

    Default `DistributedSampler` allows us to build a sampler for a dataset in DDP mode.
    Usually we can easily replace default `SequentialSampler` [when ``DataLoader(shuffle=False, ...)``] and
    `RandomSampler` [when ``DataLoader(shuffle=True, ...)``] with `DistributedSampler`. But for the custom sampler,
    we need an extra wrapper.

    Thus, this wrapper distributes indices produced by sampler among several devices for further usage.

    """

    def __init__(
        self, sampler: TAllSamplers, shuffle_samples_between_gpus: bool = True, pad_data_to_num_gpus: bool = True
    ):
        """
        Args:
            sampler: Sequential or batch sampler
            pad_data_to_num_gpus: When using DDP we should manage behavior with the last batch, because each device
                should have the same amount of data. If the sampler length is not evenly divisible by the number of
                devices, we must duplicate part of the data (``pad_data_to_num_gpus=True``), or discard part of the
                data (``pad_data_to_num_gpus=False``).
            shuffle_samples_between_gpus: shuffle available indices before feeding them to GPU. Note, that shuffle
                inside GPU after the feeding will be used according to behavior of the sampler.

        Note: Wrapper can be used with both the default `SequentialSampler` or `RandomSampler` from `PyTorch` and
        with some custom sampler.

        """
        super().__init__(
            dataset=_Sampler2Dataset(sampler), shuffle=shuffle_samples_between_gpus, drop_last=not pad_data_to_num_gpus
        )
        self.seed_shift_per_epoch = 0
        self.sampler = sampler

    def _reload(self) -> None:
        """
        We need to re-instantiate the wrapper in order to update the available indices for the new epoch.
        We don't perform this step on the epoch 0, because we want to be comparable with no DDP setup there.

        """
        if self.seed_shift_per_epoch > 0:
            super().__init__(dataset=_Sampler2Dataset(self.sampler), shuffle=self.shuffle, drop_last=self.drop_last)
            self.set_epoch(self.seed_shift_per_epoch)
        self.seed_shift_per_epoch += 1

    def __iter__(self) -> TAllSamplers:
        self._reload()

        for sampler_idx in super().__iter__():
            yield self.dataset[sampler_idx]


def extract_loader_parameters(loader: DataLoader, ignore_data_related_parameters: bool = False) -> Dict[str, Any]:
    """
    The function extracts parameters from dataloader, such as `collate_fn`, `num_workers`, etc, and automatically
    handles some new parameters, e.g. `prefetch_factor`.

    Args:
        loader: loader from which parameters are extracted
        ignore_data_related_parameters: The flag allows you to ignore parameters, related to data, batch content,
        and samplers.

    """

    ignore_fields = ["self"]

    if ignore_data_related_parameters:
        ignore_fields.extend(["sampler", "batch_sampler", "drop_last", "shuffle", "batch_size", "dataset"])

    extracted = {}

    signature = inspect.signature(DataLoader.__init__)
    for parameter in signature.parameters:
        if parameter not in ignore_fields:
            if hasattr(loader, parameter):
                extracted[parameter] = getattr(loader, parameter)

    assert len(extracted)

    return extracted


def patch_dataloader_to_ddp(loader: DataLoader) -> DataLoader:
    """
    Function inspects loader and modifies sampler for working in DDP mode.

    Note:
        We ALWAYS use the padding of samples (in terms of the number of batches or number of samples per epoch) in
        order to use the same amount of data for each device in DDP. Thus, the behavior with and without DDP may be
        slightly different (e.g. metrics values).

    """
    if is_ddp():
        kwargs_loader = extract_loader_parameters(loader, ignore_data_related_parameters=True)

        # If you don't spectify batch_sampler, PyTorch automatically creates default BatchSampler. In this case we
        # need convert to DDP only sampler (your custom sampler / default SequentialSampler or RandomSampler, which
        # PyTorch creates if sampler=None). We don't use `isinstance(...)` for `if` statement because we need exactly
        # class BatchSampler, ignoring any inheritance
        if type(loader.batch_sampler) is BatchSampler:
            ddp_sampler = DDPSamplerWrapper(
                sampler=loader.sampler, shuffle_samples_between_gpus=False, pad_data_to_num_gpus=True
            )
            patched_loader = DataLoader(
                dataset=loader.dataset,
                sampler=ddp_sampler,
                batch_size=loader.batch_size,
                drop_last=loader.drop_last,
                **kwargs_loader,
            )
            sampler_info = f"'{loader.sampler.__class__.__name__}' sampler"
        else:
            ddp_sampler = DDPSamplerWrapper(
                sampler=loader.batch_sampler, shuffle_samples_between_gpus=False, pad_data_to_num_gpus=True
            )
            patched_loader = DataLoader(dataset=loader.dataset, batch_sampler=ddp_sampler, **kwargs_loader)
            sampler_info = f"'{loader.batch_sampler.__class__.__name__}' batch sampler"

        logging.info(f"DataLoader with {sampler_info} is updated to DDP mode")
        return patched_loader
    else:
        warnings.warn(patch_dataloader_to_ddp.__name__, WarningDDP)
        return loader


def check_loaders_is_patched(loaders: Union[DataLoader, Sequence[DataLoader]]) -> bool:
    loaders = [loaders] if isinstance(loaders, DataLoader) else loaders

    for loader in loaders:
        if not any(isinstance(sampler, DDPSamplerWrapper) for sampler in [loader.batch_sampler, loader.sampler]):
            return False

    return True


__all__ = [
    "DDPSamplerWrapper",
    "patch_dataloader_to_ddp",
    "extract_loader_parameters",
    "check_loaders_is_patched",
]



from copy import deepcopy

from oml.interfaces.retrieval import IRetrievalPostprocessor
from oml.retrieval.retrieval_results import RetrievalResults
from oml.utils.misc_torch import AvgOnline


class ConstantThresholding(IRetrievalPostprocessor):
    def __init__(self, th: float):
        """
        Args:
            th: Distance threshold to limit the `RetrievalResults`.
        """
        self.th = th

    def process(self, rr: RetrievalResults) -> RetrievalResults:  # type: ignore
        if rr.is_empty():
            return rr.deepcopy()

        distances_upd = []
        rids_upd = []

        for dists, rids in zip(rr.distances, rr.retrieved_ids):
            mask = dists < self.th
            dists = dists[mask]
            rids = rids[mask]

            distances_upd.append(dists)
            rids_upd.append(rids)

        rr_upd = RetrievalResults(distances=distances_upd, retrieved_ids=rids_upd, gt_ids=deepcopy(rr.gt_ids))

        return rr_upd


class AdaptiveThresholding(IRetrievalPostprocessor):
    def __init__(self, n_std: float):
        """
        This postprocessor cuts `RetrievalResults` if a big gap in consecutive distances has been found.
        The big gap is defined as a gap greater than `n_std * avg_gap`.

        Args:
            n_std: the smaller value, the less `RetrievalResults` will be remained.

        """
        self.n_std = n_std

    def process(self, rr: RetrievalResults) -> RetrievalResults:  # type: ignore
        if rr.is_empty():
            return rr.deepcopy()

        avg_diff = AvgOnline()
        for dists in rr.distances:
            avg_diff.update(dists[1:] - dists[:-1])

        distances_upd = []
        rids_upd = []

        for dists, rids in zip(rr.distances, rr.retrieved_ids):
            diffs = dists[1:] - dists[:-1]
            mask_gaps = diffs > self.n_std * avg_diff.result

            if mask_gaps.sum() == 0:
                distances_upd.append(dists)
                rids_upd.append(rids)
            else:
                i_th = mask_gaps.nonzero()[0]
                distances_upd.append(dists[: i_th + 1])
                rids_upd.append(rids[: i_th + 1])

        rr_upd = RetrievalResults(distances=distances_upd, retrieved_ids=rids_upd, gt_ids=deepcopy(rr.gt_ids))

        return rr_upd


__all__ = ["AdaptiveThresholding", "ConstantThresholding"]



import hashlib
from io import BytesIO, FileIO
from pathlib import Path
from typing import List, Optional, Union

import gdown
import requests
import validators

from oml.const import CKPT_SAVE_ROOT, REQUESTS_TIMEOUT


def calc_hash(source: Union[Path, str, bytes]) -> str:
    if isinstance(source, (Path, str)):
        source = Path(source).resolve()
        descriptor = FileIO(str(source), mode="rb")
    elif isinstance(source, bytes):
        descriptor = BytesIO(source)  # type: ignore
    else:
        raise TypeError("Not supported type")
    descriptor.seek(0)
    hash_md5 = hashlib.md5()
    for chunk in iter(lambda: descriptor.read(4096), b""):
        hash_md5.update(chunk)
    return hash_md5.hexdigest()


def calc_folder_hash(folder: Union[Path, str]) -> str:
    """
    The function calculates the hash of the folder iterating over files sorted by their names.
    The function also pays attention to filenames, not only content.
    """

    folder = Path(folder)
    assert folder.exists() and folder.is_dir()
    files = sorted(tuple(str(fname.relative_to(folder).as_posix()) for fname in folder.rglob("*") if fname.is_file()))

    folder_hash = hashlib.md5()
    structure_hash = "".join(files).encode()
    folder_hash.update(structure_hash)

    for file in files:
        file_hash = calc_hash(folder / file).encode()
        folder_hash.update(file_hash)
    return folder_hash.hexdigest()


def check_exists_and_validate_md5(path: Union[str, Path], md5: Optional[str] = None) -> bool:
    path = Path(path)

    if not path.exists():
        return False

    if not md5:
        return True

    if path.is_dir():
        existing_hash = calc_folder_hash(path)
    else:
        existing_hash = calc_hash(path)

    print(f"Existing hash: {existing_hash}, expected hash: {md5}")

    return existing_hash == md5


def download_file_from_url(url: str, fname: Optional[str] = None, timeout: float = REQUESTS_TIMEOUT) -> Optional[bytes]:
    assert validators.url(url), "Invalid URL"
    response = requests.get(url, timeout=timeout)

    if response.status_code == 200:
        if fname is not None:
            Path(fname).parent.mkdir(parents=True, exist_ok=True)
            with open(fname, "wb+") as f:
                f.write(response.content)
            return None
        else:
            return response.content
    else:
        raise RuntimeError(f"Can not download file from '{url}'")


def download_checkpoint_one_of(
    url_or_fid_list: Union[List[str], str], hash_md5: str, fname: Optional[str] = None
) -> str:
    """
    The function iteratively tries to download a checkpoint from the list of resources and stops at the first
    one available for download.
    """
    if not isinstance(url_or_fid_list, (tuple, list)):
        url_or_fid_list = [url_or_fid_list]

    attempt = 0
    for url_or_fid in url_or_fid_list:
        attempt += 1
        print(url_or_fid)
        try:
            return download_checkpoint(url_or_fid, hash_md5, fname)
        except Exception:
            if attempt == len(url_or_fid_list):
                raise

    return None  # type: ignore


def download_checkpoint(url_or_fid: str, hash_md5: str, fname: Optional[str] = None) -> str:
    """
    Args:
        url_or_fid: URL to the checkpoint or file id on Google Drive
        hash_md5: Value of md5sum
        fname: Name of the checkpoint after the downloading process

    Returns:
        Path to the checkpoint

    """
    CKPT_SAVE_ROOT.mkdir(exist_ok=True, parents=True)

    fname = fname if fname else Path(url_or_fid).name
    save_path = str(CKPT_SAVE_ROOT / fname)

    if Path(save_path).exists():
        actual_hash = calc_hash(save_path)
        if actual_hash.startswith(hash_md5):
            print("Checkpoint is already here.")
            return save_path
        else:
            print(
                f"Checkpoint is already here, but hashes don't match. "
                f"{actual_hash} != {hash_md5}."
                f"We will remove the old checkpoint: "
                f"{save_path}."
            )
            Path(save_path).unlink()

    print("Downloading checkpoint...")

    if validators.url(url_or_fid):
        download_file_from_url(url=url_or_fid, fname=save_path)
    else:  # we assume we work with file id (Google Drive)
        gdown.download(id=url_or_fid, output=save_path, quiet=False)

    if not calc_hash(save_path).startswith(hash_md5):
        raise Exception("Downloaded checkpoint is probably broken. " "Hash values don't match.")

    return str(save_path)


__all__ = [
    "calc_hash",
    "calc_folder_hash",
    "check_exists_and_validate_md5",
    "download_file_from_url",
    "download_checkpoint",
    "download_checkpoint_one_of",
]



from pathlib import Path
from typing import Optional, Union

import torch

from oml.const import CKPT_SAVE_ROOT
from oml.interfaces.models import IExtractor
from oml.models.utils import (
    remove_criterion_in_state_dict,
    remove_prefix_from_state_dict,
)
from oml.models.vit_unicom.external import vision_transformer
from oml.models.vit_unicom.external.model import load  # type: ignore
from oml.utils.misc_torch import normalise


def unicom_vitb32(using_checkpoint: bool = True) -> vision_transformer.VisionTransformer:  # type: ignore
    return vision_transformer.build_model("ViT-B/32", using_checkpoint=using_checkpoint)  # type: ignore


def unicom_vitb16(using_checkpoint: bool = True) -> vision_transformer.VisionTransformer:  # type: ignore
    return vision_transformer.build_model("ViT-B/16", using_checkpoint=using_checkpoint)  # type: ignore


def unicom_vitl14(using_checkpoint: bool = True) -> vision_transformer.VisionTransformer:  # type: ignore
    return vision_transformer.build_model("ViT-L/14", using_checkpoint=using_checkpoint)  # type: ignore


def unicom_vitl14_336px(using_checkpoint: bool = True) -> vision_transformer.VisionTransformer:  # type: ignore
    return vision_transformer.build_model("ViT-L/14@336px", using_checkpoint=using_checkpoint)  # type: ignore


class ViTUnicomExtractor(IExtractor):
    constructors = {
        "vitb32_unicom": unicom_vitb32,
        "vitb16_unicom": unicom_vitb16,
        "vitl14_unicom": unicom_vitl14,
        "vitl14_336px_unicom": unicom_vitl14_336px,
    }

    pretrained_models = {
        "vitb32_unicom": {
            "download_fn": lambda: load("ViT-B/32", download_root=CKPT_SAVE_ROOT),
            "init_args": {"arch": "vitb32_unicom", "normalise_features": True},
        },
        "vitb16_unicom": {
            "download_fn": lambda: load("ViT-B/16", download_root=CKPT_SAVE_ROOT),
            "init_args": {"arch": "vitb16_unicom", "normalise_features": True},
        },
        "vitl14_unicom": {
            "download_fn": lambda: load("ViT-L/14", download_root=CKPT_SAVE_ROOT),
            "init_args": {"arch": "vitl14_unicom", "normalise_features": True},
        },
        "vitl14_336px_unicom": {
            "download_fn": lambda: load("ViT-L/14@336px", download_root=CKPT_SAVE_ROOT),
            "init_args": {"arch": "vitl14_336px_unicom", "normalise_features": True},
        },
    }

    def __init__(
        self, weights: Optional[Union[Path, str]], arch: str, normalise_features: bool, use_gradiend_ckpt: bool = True
    ):
        """
        Args:
            weights: Path to weights or a special key to download pretrained checkpoint, use ``None`` to
             randomly initialize model's weights. You can check the available pretrained checkpoints
             in ``self.pretrained_models``.
            arch: Might be one of ``vitb32_unicom``, ``vitb16_unicom``, ``vitl14_unicom``, ``vitl14_336px_unicom``.
             You can check all the available options in ``self.constructors``
            normalise_features: Set ``True`` to normalise output features
            use_gradiend_ckpt: Whether to use gradient checkpointing inside VisionTransformer class.
        """
        assert arch in self.constructors
        super(IExtractor, self).__init__()

        self.arch = arch
        self.normalise_features = normalise_features

        self.model = self.constructors[arch](using_checkpoint=use_gradiend_ckpt)

        if weights is None:
            return
        elif weights in self.constructors:
            self.model, _ = self.pretrained_models[weights]["download_fn"]()  # type: ignore
        else:
            ckpt = torch.load(weights, map_location="cpu")
            state_dict = ckpt["state_dict"] if "state_dict" in ckpt else ckpt
            state_dict = remove_criterion_in_state_dict(state_dict)
            ckpt = remove_prefix_from_state_dict(state_dict, trial_key="norm.bias")
            self.model.load_state_dict(ckpt, strict=True)

    @property
    def feat_dim(self) -> int:
        return {"vitb32_unicom": 512, "vitb16_unicom": 768, "vitl14_unicom": 768, "vitl14_336px_unicom": 768}[self.arch]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.model(x)

        if self.normalise_features:
            x = normalise(x)

        return x


__all__ = ["ViTUnicomExtractor"]



from oml.lightning.callbacks.metric import MetricValCallback
from oml.lightning.modules.extractor import ExtractorModule, ExtractorModuleDDP
from oml.lightning.modules.pairwise_postprocessing import (
    PairwiseModule,
    PairwiseModuleDDP,
)
from oml.lightning.pipelines import logging



import math
from collections import Counter
from typing import Dict, Iterator, List, Union

import numpy as np

from oml.interfaces.samplers import IBatchSampler
from oml.utils.misc import smart_sample


class CategoryBalanceSampler(IBatchSampler):
    """
    This sampler takes ``n_instances`` for each of the ``n_labels`` for each of the
    ``n_categories`` to form the batches.
    Thus, the batch size is ``n_instances x n_labels x n_categories``.

    Note, to form an epoch of batches we simply sample ``L / n_labels`` batches with repetition.

    """

    def __init__(
        self,
        labels: Union[List[int], np.ndarray],
        label2category: Dict[int, Union[str, int]],
        n_categories: int,
        n_labels: int,
        n_instances: int,
        resample_labels: bool = False,
        weight_categories: bool = True,
    ):
        """

        Args:
            labels: Labels to sample from
            label2category: Mapping from label to category
            n_categories: The desired number of categories to sample for each batch
            n_labels: The desired number of labels to sample for each category in batch
            n_instances: The desired number of samples to sample for each label in batch
            resample_labels: If ``True`` sample with repetition otherwise, otherwise
                raise an error in case of the labels lack in any category
            weight_categories: If ``True`` sample categories for each batch with weights proportional
                to the number of unique labels in the categories

        """
        unique_labels = set(labels)
        unique_categories = set(label2category.values())
        category2labels = {
            category: {label for label, cat in label2category.items() if category == cat}
            for category in unique_categories
        }
        for param in [n_categories, n_labels, n_instances]:
            if not isinstance(param, int):
                raise TypeError(f"{param.__name__} must be int, {type(param)} given")
        if not 1 <= n_categories <= len(unique_categories):
            raise ValueError(f"must be 1 <= n_categories <= {len(unique_categories)}, {n_categories} given")
        if not 1 < n_labels <= len(unique_labels):
            raise ValueError(f"must be 1 < n_labels <= {len(unique_labels)}, {n_labels} given")
        if n_instances <= 1:
            raise ValueError(f"must be not less than 1, {n_instances} given")
        if any(label not in label2category.keys() for label in unique_labels):
            raise ValueError("All the labels must have category")
        if any(label not in unique_labels for label in label2category.keys()):
            raise ValueError("All the labels from label2category mapping must be in the labels")
        if any(n <= 1 for n in Counter(labels).values()):
            raise ValueError("Each class must contain at least 2 instances to fit")
        if not resample_labels:
            if any(len(list(labs)) < n_labels for labs in category2labels.values()):
                raise ValueError(f"All the categories must have at least {n_labels} unique labels")

        unique_categories = sorted(list(unique_categories))

        self._labels = np.array(labels)
        self._label2category = label2category
        self.n_categories = n_categories
        self.n_labels = n_labels
        self.n_instances = n_instances

        self._batch_size = self.n_categories * self.n_labels * self.n_instances
        self._weight_categories = weight_categories

        self._label2index = {
            label: np.arange(len(self._labels))[self._labels == label].tolist() for label in sorted(list(unique_labels))
        }
        self._category2labels = {
            category: {label for label, cat in self._label2category.items() if category == cat}
            for category in unique_categories
        }
        category_weights = {cat: len(labels) / len(unique_labels) for cat, labels in self._category2labels.items()}
        self._category_weights = (
            [category_weights[cat] for cat in unique_categories] if self._weight_categories else None
        )
        self._batch_number = math.ceil(len(unique_labels) / self.n_labels)

    @property
    def batch_size(self) -> int:
        return self._batch_size

    def __len__(self) -> int:
        return self._batch_number

    def __iter__(self) -> Iterator[List[int]]:
        epoch_indices = []
        for _ in range(self._batch_number):
            categories = np.random.choice(
                list(self._category2labels.keys()),
                size=self.n_categories,
                replace=False,
                p=self._category_weights,
            )
            batch_indices = []
            for category in categories:
                labels_available = list(self._category2labels[category])
                labels = smart_sample(array=labels_available, k=self.n_labels)
                for label in labels:
                    indices = self._label2index[label]
                    samples_indices = smart_sample(array=indices, k=self.n_instances)
                    batch_indices.extend(samples_indices)
            epoch_indices.append(batch_indices)
        return iter(epoch_indices)


__all__ = ["CategoryBalanceSampler"]



import warnings
from copy import deepcopy
from pprint import pprint
from typing import Any, Collection, Dict, Iterable, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch import FloatTensor, LongTensor

from oml.const import (
    CATEGORIES_COLUMN,
    EMBEDDINGS_KEY,
    LOG_TOPK_IMAGES_PER_ROW,
    LOG_TOPK_ROWS_PER_METRIC,
    OVERALL_CATEGORIES_KEY,
)
from oml.ddp.utils import is_main_process
from oml.functional.metrics import (
    TMetricsDict,
    calc_fnmr_at_fmr_by_distances,
    calc_retrieval_metrics,
    calc_topological_metrics,
    reduce_metrics,
)
from oml.interfaces.datasets import IQueryGalleryLabeledDataset, IVisualizableDataset
from oml.interfaces.metrics import IMetricVisualisable, TIndices
from oml.interfaces.retrieval import IRetrievalPostprocessor
from oml.metrics.accumulation import Accumulator
from oml.retrieval.retrieval_results import RetrievalResults
from oml.utils.misc import flatten_dict, pad_array_right, remove_unused_kwargs


def calc_retrieval_metrics_rr(
    rr: RetrievalResults,
    query_categories: Optional[Union[LongTensor, np.ndarray]] = None,
    cmc_top_k: Tuple[int, ...] = (5,),
    precision_top_k: Tuple[int, ...] = (5,),
    map_top_k: Tuple[int, ...] = (5,),
    reduce: bool = True,
    verbose: bool = True,
) -> TMetricsDict:
    """
    Function to compute different retrieval metrics.

    Args:
        rr: An instance of `RetrievalResults`.
        query_categories: Categories of queries with the size of ``n_query`` to compute metrics for each category.
        cmc_top_k: Values of ``k`` to calculate ``cmc@k`` (`Cumulative Matching Characteristic`)
        precision_top_k: Values of ``k`` to calculate ``precision@k``
        map_top_k: Values of ``k`` to calculate ``map@k`` (`Mean Average Precision`)
        reduce: If ``False`` return metrics for each query without averaging
        verbose: Set ``True`` to make the function verbose.

    Returns:
        Metrics dictionary.

    """
    return calc_retrieval_metrics(
        retrieved_ids=rr.retrieved_ids,
        gt_ids=rr.gt_ids,
        cmc_top_k=cmc_top_k,
        precision_top_k=precision_top_k,
        map_top_k=map_top_k,
        query_categories=query_categories,
        reduce=reduce,
        verbose=verbose,
    )


def calc_fnmr_at_fmr_rr(
    rr: RetrievalResults,
    fmr_vals: Tuple[float, ...] = (0.1,),
) -> TMetricsDict:
    """
    For more details see `calc_fnmr_at_fmr` docs.

    Args:
        rr: An instance of `RetrievalResults`.:
        fmr_vals: Values of `FMR` to calculate `FNMR` at.

    Returns:
        Metrics dictionary.
    """

    max_size = max(len(d) for d in rr.distances)
    dist = np.stack([pad_array_right(np.array(d), max_size, val=-1) for d in rr.distances])

    mask_gt = np.zeros(dist.shape, dtype=bool)
    mask_not_padding = np.ones(dist.shape, dtype=bool)

    for i, (ri, gt_id) in enumerate(zip(rr.retrieved_ids, rr.gt_ids)):
        is_correct = torch.isin(ri, gt_id)
        mask_gt[i, : len(is_correct)] = is_correct
        mask_not_padding[i, len(ri) :] = False

    pos_dist = dist[mask_gt & mask_not_padding].flatten()
    neg_dist = dist[~mask_gt & mask_not_padding].flatten()

    return calc_fnmr_at_fmr_by_distances(pos_dist=pos_dist, neg_dist=neg_dist, fmr_vals=fmr_vals)


class EmbeddingMetrics(IMetricVisualisable):
    """
    This class is designed to accumulate model outputs produced for every batch.
    Since retrieval metrics are not additive, we can compute them only after all data has been collected.

    """

    metric_name = ""

    def __init__(
        self,
        dataset: Optional[IQueryGalleryLabeledDataset],
        cmc_top_k: Tuple[int, ...] = (5,),
        precision_top_k: Tuple[int, ...] = (5,),
        map_top_k: Tuple[int, ...] = (5,),
        fmr_vals: Tuple[float, ...] = tuple(),
        pcf_variance: Tuple[float, ...] = (0.5,),
        postprocessor: Optional[IRetrievalPostprocessor] = None,
        metrics_to_exclude_from_visualization: Iterable[str] = (),
        return_only_overall_category: bool = False,
        visualize_only_overall_category: bool = True,
        verbose: bool = True,
    ):
        """

        Args:
            dataset: Annotated dataset having query-gallery split.
            cmc_top_k: Values of ``k`` to calculate ``cmc@k`` (`Cumulative Matching Characteristic`)
            precision_top_k: Values of ``k`` to calculate ``precision@k``
            map_top_k: Values of ``k`` to calculate ``map@k`` (`Mean Average Precision`)
            fmr_vals: Values of ``fmr`` (measured in quantiles) to calculate ``fnmr@fmr`` (`False Non Match Rate
                      at the given False Match Rate`).
                      For example, if ``fmr_values`` is (0.2, 0.4) we will calculate ``fnmr@fmr=0.2``
                      and ``fnmr@fmr=0.4``.
                      Note, computing this metric requires additional memory overhead,
                      that is why it's turned off by default.
            pcf_variance: Values in range [0, 1]. Find the number of components such that the amount
                          of variance that needs to be explained is greater than the percentage specified
                          by ``pcf_variance``.
            postprocessor: Postprocessor which applies some techniques like query reranking
            metrics_to_exclude_from_visualization: Names of the metrics to exclude from the visualization. It will not
             affect calculations.
            return_only_overall_category: Set ``True`` if you want to return only the aggregated metrics
            visualize_only_overall_category: Set ``False`` if you want to visualize each category separately
            verbose: Set ``True`` if you want to print metrics

        """
        self.dataset = dataset
        self.cmc_top_k = cmc_top_k
        self.precision_top_k = precision_top_k
        self.map_top_k = map_top_k
        self.fmr_vals = fmr_vals
        self.pcf_variance = pcf_variance
        self.postprocessor = postprocessor

        self.retrieval_results: Optional[RetrievalResults] = None

        self.metrics: Optional[TMetricsDict] = None
        self.metrics_unreduced: Optional[TMetricsDict] = None

        self.visualize_only_overall_category = visualize_only_overall_category
        self.return_only_overall_category = return_only_overall_category

        self.metrics_to_exclude_from_visualization = ["fnmr@fmr", "pcf", *metrics_to_exclude_from_visualization]
        self.verbose = verbose

        self._acc_embeddings_key = "__embeddings"
        self.acc = Accumulator(keys_to_accumulate=(self._acc_embeddings_key,))

        if fmr_vals:
            warnings.warn("Note, computing FNMR@FMR may significantly decrease computation time and memory consuming!")

    def setup(self, num_samples: Optional[int] = None) -> None:  # type: ignore
        self.retrieval_results = None
        self.metrics = None
        self.metrics_unreduced = None

        num_samples = num_samples if num_samples is not None else len(self.dataset)
        self.acc.refresh(num_samples)

    def update(self, embeddings: FloatTensor, indices: TIndices) -> None:
        """
        Args:
            embeddings: Representations of the dataset items containing in the current batch.
            indices: Global indices of the dataset items within the range of ``(0, dataset_size - 1)``.
                     Indices are needed to make sure that we can align dataset items and collected information.

        """
        indices = indices if isinstance(indices, List) else indices.tolist()
        self.acc.update_data(data_dict={self._acc_embeddings_key: embeddings}, indices=indices)

    def update_data(self, data: Dict[str, Any], indices: TIndices) -> Any:
        self.update(embeddings=data[EMBEDDINGS_KEY], indices=indices)

    def _compute_retrieval_results(self) -> None:
        # note, fmr requires a lot of compute
        fmr_vals = len(self.dataset.get_gallery_ids()) if self.fmr_vals else 1
        max_k = max([*self.cmc_top_k, *self.precision_top_k, *self.map_top_k, fmr_vals])
        if self.postprocessor:
            # todo: refactor how we deal with postprocessors after we have more examples
            top_n = getattr(self.postprocessor, "top_n", len(self.dataset.get_gallery_ids()))
            max_k = max(max_k, top_n)

        self.retrieval_results = RetrievalResults.from_embeddings(  # type: ignore
            embeddings=self.acc.storage[self._acc_embeddings_key],
            dataset=self.dataset,
            n_items=max_k,
            verbose=self.verbose,
        )

        if self.postprocessor:
            args = {"rr": self.retrieval_results, "dataset": self.dataset}
            args = remove_unused_kwargs(args, self.postprocessor.process)
            self.retrieval_results = self.postprocessor.process(**args)

    def compute_metrics(self) -> TMetricsDict:  # type: ignore
        self.acc = self.acc.sync()  # gathering data from devices happens here if DDP

        if not self.acc.is_storage_full():
            raise ValueError(
                f"Metrics have to be calculated on fully collected data. "
                f"The size of the current storage is less than num samples: "
                f"we've collected {self.acc.collected_samples} out of {self.acc.num_samples}."
            )

        self._compute_retrieval_results()

        args_r = {
            "cmc_top_k": self.cmc_top_k,
            "precision_top_k": self.precision_top_k,
            "map_top_k": self.map_top_k,
            "rr": self.retrieval_results,
            "reduce": False,
            "verbose": self.verbose,
        }

        args_t = {"embeddings": self.acc.storage[self._acc_embeddings_key], "pcf_variance": self.pcf_variance}

        if CATEGORIES_COLUMN in self.dataset.extra_data:
            categories = np.array(self.dataset.extra_data[CATEGORIES_COLUMN])
            query_categories = categories[self.dataset.get_query_ids()]

            metrics_r = calc_retrieval_metrics_rr(query_categories=query_categories, **args_r)  # type: ignore
            metrics_t = calc_topological_metrics(categories=categories, **args_t)  # type: ignore
            self.metrics_unreduced = {cat: {**metrics_r[cat], **metrics_t[cat]} for cat in metrics_r.keys()}

        else:
            metrics_r = calc_retrieval_metrics_rr(**args_r)  # type: ignore
            metrics_t = calc_topological_metrics(**args_t)  # type: ignore
            self.metrics_unreduced = {OVERALL_CATEGORIES_KEY: {**metrics_r, **metrics_t}}

        self.metrics_unreduced[OVERALL_CATEGORIES_KEY].update(
            calc_fnmr_at_fmr_rr(self.retrieval_results, self.fmr_vals)
        )

        self.metrics = reduce_metrics(deepcopy(self.metrics_unreduced))

        if self.return_only_overall_category:
            metric_to_return = {OVERALL_CATEGORIES_KEY: deepcopy(self.metrics[OVERALL_CATEGORIES_KEY])}
        else:
            metric_to_return = deepcopy(self.metrics)

        if self.verbose and is_main_process():
            print("\nMetrics:")
            pprint(metric_to_return)

        return metric_to_return

    def ready_to_visualize(self) -> bool:
        return isinstance(self.dataset, IVisualizableDataset)

    def visualize(self) -> Tuple[Collection[plt.Figure], Collection[str]]:
        """
        Visualize worst queries by all the available metrics.
        """
        metrics_flat = flatten_dict(self.metrics, ignored_keys=self.metrics_to_exclude_from_visualization)
        figures = []
        titles = []
        for metric_name in metrics_flat:
            if self.visualize_only_overall_category and not metric_name.startswith(OVERALL_CATEGORIES_KEY):
                continue
            fig = self.get_plot_for_worst_queries(
                metric_name=metric_name, n_queries=LOG_TOPK_ROWS_PER_METRIC, n_instances=LOG_TOPK_IMAGES_PER_ROW
            )
            log_str = f"top {LOG_TOPK_ROWS_PER_METRIC} worst by {metric_name}".replace("/", "_")
            figures.append(fig)
            titles.append(log_str)
        return figures, titles

    def get_worst_queries_ids(self, metric_name: str, n_queries: int) -> List[int]:
        metric_values = flatten_dict(self.metrics_unreduced)[metric_name]  # type: ignore
        return torch.topk(metric_values, min(n_queries, len(metric_values)), largest=False)[1].tolist()

    def get_plot_for_worst_queries(
        self, metric_name: str, n_queries: int, n_instances: int, verbose: bool = False
    ) -> plt.Figure:
        query_ids = self.get_worst_queries_ids(metric_name=metric_name, n_queries=n_queries)
        return self.get_plot_for_queries(query_ids=query_ids, n_instances=n_instances, verbose=verbose)

    def get_plot_for_queries(self, query_ids: List[int], n_instances: int, verbose: bool = True) -> plt.Figure:
        """
        Args:
            query_ids: Indices of the queries
            n_instances: Amount of the retrieved items to show
            verbose: Set ``True`` for additional information

        """
        assert self.retrieval_results is not None, "We are not ready to plot, because there are no retrieval results."
        assert self.metrics_unreduced is not None, "We are not ready to plot, because metrics were not calculated yet."

        fig = self.retrieval_results.visualize(
            query_ids=query_ids, n_galleries_to_show=n_instances, verbose=verbose, dataset=self.dataset
        )
        fig.tight_layout()
        return fig


__all__ = ["EmbeddingMetrics", "calc_retrieval_metrics_rr", "calc_fnmr_at_fmr_rr"]



from typing import List, Union

from torch import Tensor
from torch.nn import Module


class ITripletLossWithMiner(Module):
    """
    Base class for TripletLoss combined with Miner.

    """

    def forward(self, features: Tensor, labels: Union[Tensor, List[int]]) -> Tensor:
        """
        Args:
            features: Features with the shape ``[batch_size, features_dim]``
            labels: Labels with the size of ``batch_size``

        Returns:
            Loss value

        """
        raise NotImplementedError()


__all__ = ["ITripletLossWithMiner"]



from pathlib import Path
from pprint import pprint
from typing import Any, Dict, Tuple

import pytorch_lightning as pl
from torch.utils.data import DataLoader

from oml.const import TCfg
from oml.datasets.images import get_retrieval_images_datasets
from oml.lightning.callbacks.metric import MetricValCallback
from oml.lightning.modules.extractor import ExtractorModule, ExtractorModuleDDP
from oml.lightning.pipelines.parser import (
    check_is_config_for_ddp,
    parse_engine_params_from_config,
)
from oml.metrics.embeddings import EmbeddingMetrics
from oml.registry.models import get_extractor_by_cfg
from oml.registry.postprocessors import get_postprocessor_by_cfg
from oml.registry.transforms import get_transforms_by_cfg
from oml.utils.misc import dictconfig_to_dict


def extractor_validation_pipeline(cfg: TCfg) -> Tuple[pl.Trainer, Dict[str, Any]]:
    """
    This pipeline allows you to validate a feature extractor which
    represents images as feature vectors.

    The config can be specified as a dictionary or with hydra: https://hydra.cc/.
    For more details look at ``pipelines/features_extraction/README.md``

    """
    cfg = dictconfig_to_dict(cfg)
    trainer_engine_params = parse_engine_params_from_config(cfg)
    is_ddp = check_is_config_for_ddp(trainer_engine_params)

    pprint(cfg)

    _, valid_dataset = get_retrieval_images_datasets(
        dataset_root=Path(cfg["dataset_root"]),
        transforms_train=None,
        transforms_val=get_transforms_by_cfg(cfg["transforms_val"]),
        dataframe_name=cfg["dataframe_name"],
    )
    loader_val = DataLoader(dataset=valid_dataset, batch_size=cfg["bs_val"], num_workers=cfg["num_workers"])

    extractor = get_extractor_by_cfg(cfg["extractor"])

    module_kwargs = {}
    if is_ddp:
        module_kwargs["loaders_val"] = loader_val
        module_constructor = ExtractorModuleDDP
    else:
        module_constructor = ExtractorModule  # type: ignore

    pl_model = module_constructor(
        extractor=extractor,
        criterion=None,
        optimizer=None,
        scheduler=None,
        input_tensors_key=valid_dataset.input_tensors_key,
        labels_key=valid_dataset.labels_key,
        **module_kwargs,
    )

    postprocessor = None if not cfg.get("postprocessor", None) else get_postprocessor_by_cfg(cfg["postprocessor"])

    # Note! We add the link to our extractor to a Lightning's Module, so it can recognize it and manipulate its devices
    # todo: refactor it later after we have more examples of postprocessors
    if (postprocessor is not None) and hasattr(postprocessor, "model"):
        pl_model.model_link_ = postprocessor.model  # type: ignore

    metrics_calc = EmbeddingMetrics(
        dataset=valid_dataset,
        postprocessor=postprocessor,
        **cfg.get("metric_args", {}),
    )
    clb_metric = MetricValCallback(metric=metrics_calc, log_images=False)

    trainer = pl.Trainer(callbacks=[clb_metric], precision=cfg.get("precision", 32), **trainer_engine_params)

    if is_ddp:
        logs = trainer.validate(verbose=True, model=pl_model)
    else:
        logs = trainer.validate(dataloaders=loader_val, verbose=True, model=pl_model)

    return trainer, logs


__all__ = ["extractor_validation_pipeline"]



from abc import ABC
from collections.abc import MutableMapping
from contextlib import contextmanager
from typing import Any, Dict, Hashable, Iterator, List, Optional, Tuple, Type, Union

import numpy as np
import torch
from torch import Tensor, cdist

TSingleValues = Union[int, float, np.float64, np.int_, torch.Tensor]
TSequenceValues = Union[List[float], Tuple[float, ...], np.ndarray, torch.Tensor]
TOnlineValues = Union[TSingleValues, TSequenceValues]


def take_2d(x: Tensor, indices: Tensor) -> Tensor:
    """
    Args:
        x: Tensor with the shape of ``[N, M]``
        indices: Tensor of integers with the shape of ``[N, P]``
            Note, rows in ``indices`` may contain duplicated values.
            It means that we can take the same element from ``x`` several times.

    Returns:
        Tensor of the items picked from ``x`` with the shape of ``[N, P]``

    """
    assert x.ndim == indices.ndim == 2
    assert x.shape[0] == indices.shape[0]

    n = x.shape[0]
    ii = torch.arange(n).unsqueeze(-1).expand(n, indices.shape[1])

    return x[ii, indices]


def assign_2d(x: Tensor, indices: Tensor, new_values: Tensor) -> Tensor:
    """
    Args:
        x: Tensor with the shape of ``[N, M]``
        indices: Tensor of integers with the shape of ``[N, P]``, where ``P <= M``
        new_values: Tensor with the shape of ``[N, P]``

    Returns:
        Tensor with the shape of ``[N, M]`` constructed by the following rule:
        ``x[i, indices[i, j]] = new_values[i, j]``

    """
    assert x.ndim == indices.ndim == new_values.ndim
    assert x.shape[0] == indices.shape[0] == new_values.shape[0]
    assert indices.shape == new_values.shape

    n = x.shape[0]
    ii = torch.arange(n).unsqueeze(-1).expand(n, indices.shape[1])

    x[ii, indices] = new_values

    return x


def cat_two_sorted_tensors_and_keep_it_sorted(x1: Tensor, x2: Tensor, eps: float = 1e-6) -> Tensor:
    """
    Args:
        x1: Sorted tensor with the shape of ``[N, M]``
        x2: Sorted tensor with the shape of ``[N, P]``
        eps: Eps to have a gap between the last x1 and the first x2

    Returns:
        Concatenation of two sorted tensors.
        The first tensor may be rescaled if needed to keep the order sorted.

    """
    assert eps >= 0

    if not x1.numel():
        return x2

    if not x2.numel():
        return x1

    assert x1.shape[0] == x2.shape[0], (x1.shape[0], x2.shape[0])

    scale = (x2[:, 0] / x1[:, -1]).view(-1, 1).type_as(x1)
    need_scaling = x1[:, -1] > x2[:, 0]
    x1[need_scaling] = x1[need_scaling] * scale[need_scaling] - eps

    x = torch.concatenate([x1, x2], dim=1).float()

    return x


def elementwise_dist(x1: Tensor, x2: Tensor, p: int = 2) -> Tensor:
    """
    Args:
        x1: tensor with the shape of [N, D]
        x2: tensor with the shape of [N, D]
        p: degree

    Returns: elementwise distances with the shape of [N]

    """
    assert len(x1.shape) == len(x2.shape) == 2
    assert x1.shape == x2.shape

    # we need an extra dim here to avoid pairwise behavior of torch.cdist
    if len(x1.shape) == 2:
        x1 = x1.unsqueeze(1)
        x2 = x2.unsqueeze(1)

    dist = cdist(x1=x1, x2=x2, p=p).view(len(x1))

    return dist


def pairwise_dist(x1: Tensor, x2: Tensor, p: int = 2) -> Tensor:
    """
    Args:
        x1: tensor with the shape of [N, D]
        x2: tensor with the shape of [M, D]
        p: degree

    Returns: pairwise distances with the shape of [N, M]

    """
    assert len(x1.shape) == len(x2.shape) == 2
    assert x1.shape[-1] == x2.shape[-1]

    return cdist(x1=x1, x2=x2, p=p)


def normalise(x: Tensor, p: int = 2) -> Tensor:
    """
    Args:
        x: A 2D tensor
        p: Specifies the exact p-norm

    Returns:
        Normalised input

    """
    assert x.ndim == 2
    xn = torch.linalg.norm(x, p, dim=1).detach()
    x = x.div(xn.unsqueeze(1))
    return x


def get_device(model: torch.nn.Module) -> str:
    return str(next(model.parameters()).device)


def is_sorted_tensor(x: Tensor) -> bool:
    """
    Here is the robust checker if a tensor is sorted or not which handles case
    where some elements of the tensors are close to each other than computational tolerance.

    If we cannot distinguish elements in the tensor we assume it's sorted by default.
    """
    assert x.ndim == 1
    if len(x) == 1:
        return True

    diffs = x[1:] - x[:-1]
    diffs[torch.abs(diffs) < 10 * torch.finfo(x.dtype).eps] = 0

    return (diffs >= 0).all()


def _check_is_sequence(val: Any) -> bool:
    try:
        len(val)
        return True
    except Exception:
        return False


TData = Tuple[List[Any], Tensor, np.ndarray]


def unique_by_ids(ids: List[int], data: TData) -> Tuple[List[int], TData]:
    """
    The function sort data by the corresponding indices and drops duplicates.
    Thus, if there are multiple occurrences of the same id, it takes the first one.

    Args:
        ids: Indices of data with the length of ``N``
        data: Data with the length of ``N``

    Returns:
        Unique data records with their ids in the sorted order without duplicates

    """
    assert len(ids) == len(data)
    assert isinstance(ids, list) and len(ids) >= 1

    ids_unq, positions_unq = np.unique(ids, return_index=True)

    if isinstance(data, (list, tuple)):
        data = [data[i] for i in positions_unq]  # type: ignore
    else:
        data = data[positions_unq]

    return ids_unq.tolist(), data


@contextmanager
def temporary_setting_model_mode(model: torch.nn.Module, set_train: bool) -> torch.nn.Module:
    prev_mode = model.training
    model.train(set_train)
    yield model
    model.train(prev_mode)


class OnlineCalc(ABC):
    """
    The base class to calculate some statistics online (on the stream of values).

    """

    def __init__(self, val: Optional[TOnlineValues] = None):
        self.result: float = 0.0
        if val is not None:
            self.update(val)

    def update(self, val: TOnlineValues) -> None:
        if _check_is_sequence(val):
            self.calc_with_batch(val)
        else:
            self.calc_with_single_value(val)

    def calc_with_single_value(self, val: TSingleValues) -> None:
        """
        Calculation with non iterable types: float, int / numpy and torch elements (array and elements of
        array have different types and methods)

        """
        self.calc_with_batch([val])

    def calc_with_batch(self, val: TSequenceValues) -> None:
        """
        Vectorized calculation of online value

        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self.__dict__)})"


class AvgOnline(OnlineCalc):
    def __init__(self, *args: Any, **kwargs: Any):
        self.n = 0
        super().__init__(*args, **kwargs)

    def calc_with_batch(self, val: TSequenceValues) -> None:
        len_val = len(val)
        self.n += len_val
        self.result = sum(val) / self.n + (self.n - len_val) / self.n * self.result  # type: ignore


class SumOnline(OnlineCalc):
    def calc_with_batch(self, val: TSequenceValues) -> None:
        self.result += float(sum(val))


class OnlineDict(MutableMapping):  # type: ignore
    """
    We don't inherit from built-in 'dict' due to internal C optimization. We mimic to dict with MutableMapping
    https://treyhunner.com/2019/04/why-you-shouldnt-inherit-from-list-and-dict-in-python/

    """

    online_calculator: Type[OnlineCalc]

    def __init__(self, input_dict: Optional[Dict[Hashable, TOnlineValues]] = None):
        self.dict: Dict[Hashable, OnlineCalc] = {}

        if input_dict:
            self.update(input_dict)

    def __setitem__(self, key: Hashable, value: TOnlineValues) -> None:
        self.dict[key] = self.online_calculator(value)

    def __getitem__(self, key: Hashable) -> float:
        return self.dict[key].result

    def __len__(self) -> int:
        return len(self.dict)

    def __iter__(self) -> Iterator[Any]:
        return iter(self.dict)

    def __delitem__(self, key: Hashable) -> None:
        return self.dict.__delitem__(key)

    def update(self, input_dict: Dict[Hashable, TOnlineValues]) -> None:  # type: ignore
        for k, v in input_dict.items():
            if k in self.dict:
                self.dict[k].update(v)
            else:
                self[k] = v

    def __repr__(self) -> str:
        output = {k: v.result for k, v in self.dict.items()}
        return f"{self.__class__.__name__}({output})"

    def get_dict_with_results(self) -> Dict[str, float]:
        return {k: v for k, v in self.items()}


class OnlineAvgDict(OnlineDict):
    online_calculator = AvgOnline


class OnlineSumDict(OnlineDict):
    online_calculator = SumOnline


class PCA:
    """
    Principal component analysis (PCA).

    Estimate principal axes, and perform vectors transformation.

    Note:
        The code is almost the same as one from sklearn, but we had two reasons to have our own implementation.
        First, we need to work with Torch tensors instead of NumPy arrays. Second, we wanted to avoid one more external
        dependency.

    Attributes:
        components: Matrix of shape ``[embeddings_dim, embeddings_dim]``. Principal axes in embeddings space,
            representing the directions of maximum variance in the data. Equivalently, the right singular
            vectors of the centered input data, parallel to its eigenvectors. The components are sorted by
            ``explained_variance``.
        explained_variance: Array of size ``embeddings_dim``
            The amount of variance explained by each of the selected components.
            The variance estimation uses ``n_embeddings - 1`` degrees of freedom.
            Equal to  eigenvalues of the covariance matrix of ``embeddings``.
        explained_variance_ratio: Array of size ``embeddings_dim``.
            Percentage of variance explained by each of the components.
        singular_values: Array of size ``embeddings_dim``.
            The singular values corresponding to each of the selected components.
        mean: Array of size ``embeddings_dim``.
            Per-feature empirical mean, estimated from the training set.
            Equal to ``embeddings.mean(dim=0)``.

    For an embeddings matrix :math:`X` of shape :math:`n\\times d` the principal axes could be found by
    performing Singular Value Decomposition

    .. math::
        X = U\\Sigma V^T

    where :math:`U` is an :math:`n\\times n` orthogonal matrix, :math:`\\Sigma` is an :math:`n\\times d` rectangular
    diagonal matrix with non-negative real numbers on the diagonal, :math:`V` is an :math:`d\\times d` orthogonal
    matrix.

    Rows of the :math:`V` form an orthonormal basis, and could be used to project embeddings to a new space, possible
    of lower dimension:

    .. math::
        X' = X\\cdot V^T

    The inverse transform is done by

    .. math::
        X = X'\\cdot V

    See:

        `Principal Components Analysis`_

    .. _`Principal Components Analysis`:
        https://en.wikipedia.org/wiki/Principal_component_analysis

    Example:
        >>> embeddings = torch.rand(100, 5)
        >>> pca = PCA(embeddings)
        >>> embeddings_transformed = pca.transform(embeddings)
        >>> embeddings_recovered = pca.inverse_transform(embeddings_transformed)
        >>> torch.all(torch.isclose(embeddings, embeddings_recovered, atol=1.e-6))
        tensor(True)
    """

    components: torch.Tensor
    mean: torch.Tensor
    singular_values: torch.Tensor
    explained_variance: torch.Tensor
    explained_variance_ratio: torch.Tensor

    def __init__(self, embeddings: torch.Tensor):
        """
        Args:
            embeddings: Embeddings matrix with the shape of ``[n_embeddings, embeddings_dim]``.

        """
        self._fit(embeddings)

    def _fit(self, embeddings: torch.Tensor) -> None:
        """
        Perform the PCA. Evaluate ``components``, ``explained_variance``, ``explained_variance_ratio``,
        ``singular_values``, ``mean``.

        Args:
            embeddings: Embeddings matrix with the shape of ``[n_embeddings, embeddings_dim]``.
        """
        # The code below mirrors code from scikit-learn repository:
        # https://github.com/scikit-learn/scikit-learn/blob/f3f51f9b6/sklearn/decomposition/_pca.py#L491
        n_samples = embeddings.shape[0]
        self.mean = embeddings.mean(dim=0).unsqueeze(0)
        embeddings = embeddings - self.mean
        # if there are more embeddings than its dimension, then we will not perform full matrices evaluation
        full_matrices = embeddings.shape[0] < embeddings.shape[1]
        _, self.singular_values, self.components = torch.linalg.svd(embeddings, full_matrices=full_matrices)
        self.explained_variance = self.singular_values**2 / (n_samples - 1)
        self.explained_variance_ratio = self.explained_variance / self.explained_variance.sum()

        # Make components deterministic.
        # Note. In sklearn this operation is done based on the U matrix of SVD decomposition, and
        # here V matrix is used. So, the components of this class and sklearn.decomposition.PCA could differ in sign.
        # See the following for details:
        # https://github.com/scikit-learn/scikit-learn/blob/f3f51f9b6/sklearn/decomposition/_pca.py#L520
        n_components = self.components.shape[0]
        max_abs_rows = torch.argmax(torch.abs(self.components), dim=1)
        signs = torch.sign(self.components[torch.arange(n_components), max_abs_rows])
        self.components *= signs.unsqueeze(1)

    def transform(self, embeddings: torch.Tensor, n_components: Optional[int] = None) -> torch.Tensor:
        """
        Apply fitted PCA to transform embeddings.

        Args:
            embeddings: Matrix of shape ``[n_embeddings, embeddings_dim]``.
            n_components: The desired dimension of the output.

        Returns:
            Transformed embeddings.
        """
        if not n_components:
            n_components = embeddings.shape[1]
        self._check_dimensions(n_components)
        embeddings_ = embeddings.to(self.mean) - self.mean
        return torch.matmul(embeddings_, self.components[:n_components, :].T).to(embeddings)

    def inverse_transform(self, embeddings: torch.Tensor) -> torch.Tensor:
        """
        Apply inverse transform to embeddings.

        Args:
            embeddings: Matrix of shape ``[n_embeddings, N]`` where ``N <= embeddings_dim`` is the dimension of
                        embeddings.

        Returns:
            Embeddings projected into original embeddings space.
        """
        n_components = embeddings.shape[1]
        self._check_dimensions(n_components)
        return torch.matmul(embeddings, self.components[:n_components, :]) + self.mean

    def calc_principal_axes_number(self, pcf_variance: Tuple[float, ...]) -> torch.Tensor:
        """
        Function estimates the number of principal axes that are required to explain the `explained_variance_ths`
        variance.

        Args:
            pcf_variance: Values in range [0, 1]. Find the number of components such that the amount
                          of variance that needs to be explained is greater than the fraction specified
                          by ``pcf_variance``.
        Returns:
            List of amount of principal axes.

        Let :math:`X` be a set of :math:`d` dimensional embeddings.
        Let :math:`\\lambda_1, \\ldots, \\lambda_d\\in\\mathbb{R}` be a set of eigenvalues
        of the covariance matrix of :math:`X` sorted in descending order.
        Then for a given value of desired explained variance :math:`r`,
        the number of principal components that explains :math:`r\\cdot 100\\%%` variance is the largest integer
        :math:`n` such that

        .. math::
            \\frac{\\sum\\limits_{i = 1}^{n - 1}\\lambda_i}{\\sum\\limits_{i = 1}^{d}\\lambda_i} \\leq r

        Example:
            In the example bellow there are 4 vectors of length 10, and only first 4 dimensions have non-zero values.
            Its covariance matrix will have only 4 eigenvalues, that are greater than 0, i.e. there are only 4 principal
            axes. So, in order to keep at least 50% of the information from the set, we need to keep 2 principal
            axes, and in order to keep all the information we need to keep 5 principal axes (one additional axis appears
            because the number of principal axes is superior to the desired explained variance threshold).

            >>> embeddings = torch.eye(4, 10, dtype=torch.float)
            >>> pca = PCA(embeddings)
            >>> pca.calc_principal_axes_number(pcf_variance=(0.5, 1))
            tensor([2, 5])

        """
        ratio_cumsum = torch.cumsum(self.explained_variance_ratio, dim=0)
        n_components = torch.searchsorted(ratio_cumsum, torch.tensor(pcf_variance), side="right") + 1
        return n_components

    def _check_dimensions(self, n_components: int) -> None:
        if n_components > self.components.shape[0]:
            raise ValueError(
                "The embeddings couldn't be transformed, due to dimensions mismatch. "
                f"Expected dimension less than or equal to {self.components.shape[0]}, but got {n_components}"
            )


__all__ = [
    "elementwise_dist",
    "cat_two_sorted_tensors_and_keep_it_sorted",
    "pairwise_dist",
    "OnlineCalc",
    "AvgOnline",
    "SumOnline",
    "OnlineDict",
    "OnlineAvgDict",
    "OnlineSumDict",
    "take_2d",
    "assign_2d",
    "PCA",
    "unique_by_ids",
    "normalise",
    "is_sorted_tensor",
]
